"""
A helper module for defining denylists for sources programmatically
(in lamens terms, this lets you remove some output from a module you don't want)

For docs, see doc/DENYLIST.md
"""

from __future__ import annotations

import functools
import json
import sys
from collections import defaultdict
from collections.abc import Iterator, Mapping
from pathlib import Path
from typing import Any, TypeVar

import click
from more_itertools import seekable

from .serialize import dumps
from .warnings import medium

T = TypeVar("T")

DenyMap = Mapping[str, set[Any]]


def _default_key_func(obj: T) -> str:
    return str(obj)


class DenyList:
    def __init__(self, denylist_file: Path | str) -> None:
        self.file = Path(denylist_file).expanduser().absolute()
        self._deny_raw_list: list[dict[str, Any]] = []
        self._deny_map: DenyMap = defaultdict(set)

        # deny cli, user can override these
        self.fzf_path = None
        self._fzf_options = ()
        self._deny_cli_key_func = None

    def _load(self) -> None:
        if not self.file.exists():
            medium(f"denylist file {self.file} does not exist")
            return

        deny_map: DenyMap = defaultdict(set)
        data: list[dict[str, Any]] = json.loads(self.file.read_text())
        self._deny_raw_list = data

        for ignore in data:
            for k, v in ignore.items():
                deny_map[k].add(v)

        self._deny_map = deny_map

    def load(self) -> DenyMap:
        self._load()
        return self._deny_map

    def write(self) -> None:
        if not self._deny_raw_list:
            medium("no denylist data to write")
            return
        self.file.write_text(json.dumps(self._deny_raw_list))

    @classmethod
    def _is_json_primitive(cls, val: Any) -> bool:
        return isinstance(val, (str, int, float, bool, type(None)))

    @classmethod
    def _stringify_value(cls, val: Any) -> Any:
        # if it's a primitive, just return it
        if cls._is_json_primitive(val):
            return val
        # otherwise, stringify-and-back so we can compare to
        # json data loaded from the denylist file
        return json.loads(dumps(val))

    @classmethod
    def _allow(cls, obj: T, deny_map: DenyMap) -> bool:
        for deny_key, deny_set in deny_map.items():
            # this should be done separately and not as part of the getattr
            # because 'null'/None could actually be a value in the denylist,
            # and the user may define behavior to filter that out
            if not hasattr(obj, deny_key):
                return False
            val = cls._stringify_value(getattr(obj, deny_key))
            # this object doesn't have have the attribute in the denylist
            if val in deny_set:
                return False
        # if we tried all the denylist keys and didn't return False,
        # then this object is allowed
        return True

    def filter(
        self,
        itr: Iterator[T],
        *,
        invert: bool = False,
    ) -> Iterator[T]:
        denyf = functools.partial(self._allow, deny_map=self.load())
        if invert:
            return filter(lambda x: not denyf(x), itr)
        return filter(denyf, itr)

    def deny(self, key: str, value: Any, *, write: bool = False) -> None:
        '''
        add a key/value pair to the denylist
        '''
        if not self._deny_raw_list:
            self._load()
        self._deny_raw({key: self._stringify_value(value)}, write=write)

    def _deny_raw(self, data: dict[str, Any], *, write: bool = False) -> None:
        self._deny_raw_list.append(data)
        if write:
            self.write()

    def _prompt_keys(self, item: T) -> str:
        import pprint

        click.echo(pprint.pformat(item))
        # TODO: extract keys from item by checking if its dataclass/NT etc.?
        resp = click.prompt("Key to deny on").strip()
        if not hasattr(item, resp):
            click.echo(f"Could not find key '{resp}' on item", err=True)
            return self._prompt_keys(item)
        return resp

    def _deny_cli_remember(
        self,
        items: Iterator[T],
        mem: dict[str, T],
    ) -> Iterator[str]:
        # user can override _deny_cli_key_func, so it's not always None, hence the ignore
        keyf = self._deny_cli_key_func or _default_key_func  # type: ignore[redundant-expr]
        # i.e., convert each item to a string, and map str -> item
        for item in items:
            key = keyf(item)
            mem[key] = item
            yield key

    def deny_cli(self, itr: Iterator[T]) -> None:
        try:
            from pyfzf import FzfPrompt
        except ImportError:
            click.echo("pyfzf is required to use the denylist cli, run 'python3 -m pip install pyfzf_iter'", err=True)
            sys.exit(1)

        # wrap in seekable so we can use it multiple times
        # progressively caches the items as we iterate over them
        sit = seekable(itr)

        prompt_continue = True

        while prompt_continue:
            # reset the iterator
            sit.seek(0)
            # so we can map the selected string from fzf back to the original objects
            memory_map: dict[str, T] = {}
            picker = FzfPrompt(executable_path=self.fzf_path, default_options="--no-multi")
            picked_l = picker.prompt(
                self._deny_cli_remember(itr, memory_map),
                "--read0",
                *self._fzf_options,
                delimiter="\0",
            )
            assert isinstance(picked_l, list)
            if picked_l:
                picked: T = memory_map[picked_l[0]]
                key = self._prompt_keys(picked)
                self.deny(key, getattr(picked, key), write=True)
                click.echo(f"Added {self._deny_raw_list[-1]} to denylist", err=True)
            else:
                click.echo("No item selected", err=True)

            prompt_continue = click.confirm("Continue?")



"""
Instagram data (uses [[https://www.instagram.com/download/request][official GDPR export]])
"""

from __future__ import annotations

import json
import re
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

from more_itertools import bucket, spy

from my.core import (
    Json,
    Paths,
    Res,
    assert_never,
    datetime_naive,
    get_files,
    make_logger,
)
from my.core.common import unique_everseen
from my.core.compat import add_note

from my.config import instagram as user_config  # isort: skip

logger = make_logger(__name__)


@dataclass
class config(user_config.gdpr):
    # paths[s]/glob to the exported zip archives
    export_path: Paths
    # TODO later also support unpacked directories?


def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


# TODO think about unifying with stuff from android.py
@dataclass(unsafe_hash=True)
class User:
    id: str
    username: str
    full_name: str


@dataclass
class _BaseMessage:
    # ugh, this is insane, but does look like it's just keeping local device time???
    # checked against a message sent on 3 June, which should be UTC+1, but timestamp seems local
    created: datetime_naive
    text: str
    thread_id: str
    # NOTE: doesn't look like there are any meaningful message ids in the export


@dataclass(unsafe_hash=True)
class _Message(_BaseMessage):
    user_id: str


@dataclass(unsafe_hash=True)
class Message(_BaseMessage):
    user: User


def _decode(s: str) -> str:
    # yeah... idk why they do that
    return s.encode('latin-1').decode('utf8')


def _entities() -> Iterator[Res[User | _Message]]:
    # it's worth processing all previous export -- sometimes instagram removes some metadata from newer ones
    # NOTE: here there are basically two options
    # - process inputs as is (from oldest to newest)
    #   this would be more stable wrt newer exports (e.g. existing thread ids won't change)
    #   the downside is that newer exports seem to have better thread ids, so might be preferable to use them
    # - process inputs reversed (from newest to oldest)
    #   the upside is that thread ids/usernames might be better
    #   the downside is that if for example the user renames, thread ids will change _a lot_, might be undesirable..
    # (from newest to oldest)
    for path in inputs():
        yield from _entitites_from_path(path)


def _entitites_from_path(path: Path) -> Iterator[Res[User | _Message]]:
    # TODO make sure it works both with plan directory
    # idelaly get_files should return the right thing, and we won't have to force ZipPath/match_structure here
    # e.g. possible options are:
    # - if packed things are detected, just return ZipPath
    # - if packed things are detected, possibly return match_structure_wrapper
    #   it might be a bit tricky because it's a context manager -- who will recycle it?
    # - if unpacked things are detected, just return the dir as it is
    #   (possibly detect them via match_structure? e.g. what if we have a bunch of unpacked dirs)
    #
    # I guess the goal for core.structure module was to pass it to other functions that expect unpacked structure
    # https://github.com/karlicoss/HPI/pull/175
    # whereas here I don't need it..
    # so for now will just implement this adhoc thing and think about properly fixing later

    personal_info = path / 'personal_information'
    if not personal_info.exists():
        # old path, used up to somewhere between feb-aug 2022
        personal_info = path / 'account_information'

    personal_info_json = personal_info / 'personal_information.json'
    if not personal_info_json.exists():
        # new path, started somewhere around april 2024
        personal_info_json = personal_info / 'personal_information' / 'personal_information.json'

    j = json.loads(personal_info_json.read_text())
    [profile] = j['profile_user']
    pdata = profile['string_map_data']
    username = pdata['Username']['value']
    full_name = _decode(pdata['Name']['value'])

    self_user = User(
        id=username,  # there is no actual id in gdpr export.. so just make up something :shrug:
        username=username,
        full_name=full_name,
    )
    yield self_user

    files = sorted(path.rglob('messages/inbox/*/message_*.json'))  # sort for more determinism
    assert len(files) > 0, path

    # parts[-2] is the directory name, which contains username/user id
    buckets = bucket(files, key=lambda p: p.parts[-2])
    conversation_to_message_files = {k: list(buckets[k]) for k in buckets}

    for conversation, message_files in conversation_to_message_files.items():

        def iter_jsons() -> Iterator[Json]:
            # messages are in files like message_1, message_2, etc -- order them by that number
            for message_file in sorted(message_files, key=lambda p: int(p.stem.split('_')[-1])):
                logger.info(f'{message_file} : processing...')
                yield json.loads(message_file.read_text())

        (first,), jsons = spy(iter_jsons())
        # title should be the same across all files, so enough to extract only first
        conversation_title = _decode(first['title'])

        # TODO older gdpr exports had 10 alnum characters?? with no relation to server id?
        m = re.fullmatch(r'(.*)_(\d+)', conversation)
        assert m is not None

        # NOTE: conversaion_username is kinda frafile
        #   e.g. if user deletes themselves there is no more username (it becomes "instagramuser")
        #   if we use older exports we might be able to figure it out though... so think about it?
        #   it also names grouped ones like instagramuserchrisfoodishblogand25others_einihreoog
        conversation_username = m.group(1)

        # NOTE: same as in android database!
        thread_v2_id = m.group(2)

        # NOTE: I'm not actually sure it's other user's id.., since it corresponds to the whole conversation
        # but I stared a bit at these ids vs database ids and can't see any way to find the correspondence :(
        # so basically the only way to merge is to actually try some magic and correlate timestamps/message texts?
        # another option is perhaps to query user id from username with some free API
        # so I feel like there is just not guaranteed way to correlate :(
        other_user = User(
            id=conversation_username,  # again, no actual ids in gdrp, so just make something up
            username=conversation_username,
            full_name=conversation_title,
        )
        yield other_user

        def _parse_message(jm: Json) -> _Message:
            content = None
            if 'content' in jm:
                content = _decode(jm['content'])
                if content.endswith(' to your message '):
                    # ugh. for some reason these contain an extra space and that messes up message merging..
                    content = content.strip()
            else:
                if (share := jm.get('share')) is not None:
                    if (share_link := share.get('link')) is not None:
                        # somewhere around 20231007, instagram removed these from gdpr links and they show up a lot in various diffs
                        share_link = share_link.replace('feed_type=reshare_chaining&', '')
                        share_link = share_link.replace('?feed_type=reshare_chaining', '')
                        share['link'] = share_link
                    if (share_text := share.get('share_text')) is not None:
                        share['share_text'] = _decode(share_text)

                photos = jm.get('photos')
                videos = jm.get('videos')
                cc = share or photos or videos
                if cc is not None:
                    content = str(cc)

            if content is None:
                # this happens e.g. on reel shares..
                # not sure what we can do properly, GPDR has literally no other info in this case
                # on android in this case at the moment we have as content ''
                # so for consistency let's do that too
                content = ''

            timestamp_ms = jm['timestamp_ms']
            sender_name = _decode(jm['sender_name'])

            user_id = other_user.id if sender_name == conversation_title else self_user.id
            return _Message(
                created=datetime.fromtimestamp(timestamp_ms / 1000),
                text=content,
                user_id=user_id,
                thread_id=thread_v2_id,
            )

        for j in jsons:
            # todo "thread_type": "Regular" ?
            for jm in reversed(j['messages']):  # in json, they are in reverse order for some reason
                try:
                    yield _parse_message(jm)
                except Exception as e:
                    add_note(e, f'^ while parsing {jm}')
                    yield e


# TODO basically copy pasted from android.py... hmm
def messages() -> Iterator[Res[Message]]:
    id2user: dict[str, User] = {}
    for x in unique_everseen(_entities):
        if isinstance(x, Exception):
            yield x
            continue
        if isinstance(x, User):
            id2user[x.id] = x
            continue
        if isinstance(x, _Message):
            try:
                user = id2user[x.user_id]
            except Exception as e:
                add_note(e, f'^ while processing {x}')
                yield e
                continue
            yield Message(
                created=x.created,
                text=x.text,
                thread_id=x.thread_id,
                user=user,
            )
            continue
        assert_never(x)



'''
VK data (exported by [[https://github.com/Totktonada/vk_messages_backup][Totktonada/vk_messages_backup]])
'''
# note: could reuse the original repo, but little point I guess since VK closed their API
import json
from collections.abc import Iterator
from dataclasses import dataclass
from datetime import datetime

import pytz

from my.config import vk_messages_backup as config
from my.core import Json, Res, Stats, datetime_aware, get_files, stat
from my.core.common import unique_everseen

# I think vk_messages_backup used this tz?
# not sure if vk actually used to return this tz in api?
TZ = pytz.timezone('Europe/Moscow')


Uid = int


@dataclass(frozen=True)
class User:
    id: Uid
    first_name: str
    last_name: str


@dataclass(frozen=True)
class Chat:
    chat_id: str
    title: str


@dataclass(frozen=True)
class Message:
    dt: datetime_aware
    chat: Chat
    id: str  # todo not sure it's unique?
    user: User
    body: str


Users = dict[Uid, User]


def users() -> Users:
    files = get_files(config.storage_path, glob='user_*.json')
    res = {}
    for f in files:
        j = json.loads(f.read_text())
        uid = j['id']
        res[uid] = User(
            id=uid,
            first_name=j['first_name'],
            last_name=j['last_name'],
        )
    return res


GROUP_CHAT_MIN_ID = 2000000000


def _parse_chat(*, msg: Json, udict: Users) -> Chat:
    # exported with newer api, peer_id is a proper identifier both for users and chats
    peer_id = msg.get('peer_id')
    if peer_id is not None:
        chat_id = peer_id
    else:
        group_chat_id = msg.get('chat_id')
        if group_chat_id is not None:
            chat_id = GROUP_CHAT_MIN_ID + group_chat_id
        else:
            chat_id = msg['user_id']

    is_group_chat = chat_id >= GROUP_CHAT_MIN_ID
    if is_group_chat:
        title = msg['title']
    else:
        user_id = msg.get('user_id') or msg.get('from_id')
        assert user_id is not None
        user = udict[user_id]
        title = f'{user.first_name} {user.last_name}'
    return Chat(
        chat_id=chat_id,
        title=title,
    )


def _parse_msg(*, msg: Json, chat: Chat, udict: Users) -> Message:
    mid = msg['id']
    md = msg['date']

    dt = datetime.fromtimestamp(md, tz=TZ)

    # todo attachments? e.g. url could be an attachment
    # todo might be forwarded?
    mb = msg.get('body')
    if mb is None:
        mb = msg.get('text')
    assert mb is not None, msg

    out = msg['out'] == 1
    if out:
        user = udict[config.user_id]
    else:
        mu = msg.get('user_id') or msg.get('from_id')
        assert mu is not None, msg
        user = udict[mu]
    return Message(
        dt=dt,
        chat=chat,
        id=mid,
        user=user,
        body=mb,
    )


def _messages() -> Iterator[Res[Message]]:
    udict = users()

    uchats = get_files(config.storage_path, glob='userchat_*.json') + get_files(config.storage_path, glob='groupchat_*.json')
    for f in uchats:
        j = json.loads(f.read_text())
        # ugh. very annoying, sometimes not possible to extract title from last message
        # due to newer api...
        # so just do in defensively until we succeed...
        chat = None
        ex = None
        for m in reversed(j):
            try:
                chat = _parse_chat(msg=m, udict=udict)
            except Exception as e:
                ex = e
                continue
        if chat is None:
            assert ex is not None
            yield ex
            continue

        for msg in j:
            try:
                yield _parse_msg(msg=msg, chat=chat, udict=udict)
            except Exception as e:
                yield e


def messages() -> Iterator[Res[Message]]:
    # seems that during backup messages were sometimes duplicated..
    yield from unique_everseen(_messages)


def stats() -> Stats:
    return {
        **stat(users),
        **stat(messages),
    }



from . import warnings

warnings.high('my.core.kompress is deprecated. Install and use "kompress" library directly in your module (see https://github.com/karlicoss/kompress )')

# in case some older user's modules aren't migrated, import all stuff from installed kompress into the scope
from kompress import *  # type: ignore[no-redef]  # complains about 'warnings' import



"""
Feedly RSS reader
"""

import json
from abc import abstractmethod
from collections.abc import Iterator, Sequence
from datetime import datetime, timezone
from pathlib import Path
from typing import Protocol

from my.core import Paths, get_files

from .common import Subscription, SubscriptionState


class config(Protocol):
    @property
    @abstractmethod
    def export_path(self) -> Paths:
        raise NotImplementedError


def make_config() -> config:
    from my.config import feedly as user_config

    class combined_config(user_config, config): ...

    return combined_config()


def inputs() -> Sequence[Path]:
    cfg = make_config()
    return get_files(cfg.export_path)


def parse_file(f: Path) -> Iterator[Subscription]:
    raw = json.loads(f.read_text())
    for r in raw:
        # err, some even don't have website..
        rid = r['id']
        website = r.get('website', rid)  # meh
        yield Subscription(
            created_at=None,
            title=r['title'],
            url=website,
            id=rid,
        )


def states() -> Iterator[SubscriptionState]:
    for f in inputs():
        dts = f.stem.split('_')[-1]
        dt = datetime.strptime(dts, '%Y%m%d%H%M%S').replace(tzinfo=timezone.utc)
        subs = list(parse_file(f))
        yield dt, subs



from my.calendar.holidays import is_holiday

from .shared_tz_config import config  # autoused fixture


def test_is_holiday() -> None:
    assert is_holiday('20190101')
    assert not is_holiday('20180601')
    assert is_holiday('20200906')  # national holiday in Bulgaria



from my.fbmessenger import export  # isort: skip  # TODO note sure if import order matters here
from . import mixin  # isort: skip  # TODO note sure if import order matters here


def messages():
    yield from mixin.messages()
    yield from export.messages()



import sys
from datetime import datetime, timedelta

import pytest
import pytz

import my.time.tz.main as tz_main
import my.time.tz.via_location as tz_via_location
from my.core import notnone
from my.core.compat import fromisoformat

from .shared_tz_config import config  # autoused fixture


def getzone(dt: datetime) -> str:
    tz = notnone(dt.tzinfo)
    return getattr(tz, 'zone')


@pytest.mark.parametrize('fast', [False, True])
def test_iter_tzs(*, fast: bool, config) -> None:
    # TODO hmm.. maybe need to make sure we start with empty config?
    config.time.tz.via_location.fast = fast

    ll = list(tz_via_location._iter_tzs())
    zones = [x.zone for x in ll]

    if fast:
        assert zones == [
            'Europe/Rome',
            'Europe/Rome',
            'Europe/Vienna',
            'Europe/Vienna',
            'Europe/Vienna',
        ]
    else:
        assert zones == [
            'Europe/Rome',
            'Europe/Rome',
            'Europe/Ljubljana',
            'Europe/Ljubljana',
            'Europe/Ljubljana',
        ]


def test_past() -> None:
    """
    Should fallback to the 'home' location provider
    """
    dt = fromisoformat('2000-01-01 12:34:45')
    dt = tz_main.localize(dt)
    assert getzone(dt) == 'America/New_York'


def test_future() -> None:
    """
    For locations in the future should rely on 'home' location
    """
    fut = datetime.now() + timedelta(days=100)
    fut = tz_main.localize(fut)
    assert getzone(fut) == 'Europe/Moscow'


def test_get_tz(config) -> None:
    # todo hmm, the way it's implemented at the moment, never returns None?
    get_tz = tz_via_location.get_tz

    # not present in the test data
    tz = get_tz(fromisoformat('2020-01-01 10:00:00'))
    assert notnone(tz).zone == 'Europe/Sofia'

    tz = get_tz(fromisoformat('2017-08-01 11:00:00'))
    assert notnone(tz).zone == 'Europe/Vienna'

    tz = get_tz(fromisoformat('2017-07-30 10:00:00'))
    assert notnone(tz).zone == 'Europe/Rome'

    tz = get_tz(fromisoformat('2020-10-01 14:15:16'))
    assert tz is not None

    on_windows = sys.platform == 'win32'
    if not on_windows:
        tz = get_tz(datetime.min)
        assert tz is not None
    else:
        # seems this fails because windows doesn't support same date ranges
        # https://stackoverflow.com/a/41400321/
        with pytest.raises(OSError):
            get_tz(datetime.min)


def test_policies() -> None:
    naive = fromisoformat('2017-07-30 10:00:00')
    assert naive.tzinfo is None  # just in case

    # actual timezone at the time
    assert getzone(tz_main.localize(naive)) == 'Europe/Rome'

    z = pytz.timezone('America/New_York')
    aware = z.localize(naive)

    assert getzone(tz_main.localize(aware)) == 'America/New_York'

    assert getzone(tz_main.localize(aware, policy='convert')) == 'Europe/Rome'

    with pytest.raises(RuntimeError):
        tz_main.localize(aware, policy='throw')



'''
Some experimental JSON parsing, basically to ensure that all data is consumed.
This can potentially allow both for safer defensive parsing, and let you know if the data started returning more data

TODO perhaps need to get some inspiration from linear logic to decide on a nice API...
'''

from __future__ import annotations

from collections import OrderedDict
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any, cast


def ignore(w, *keys):
    for k in keys:
        w[k].ignore()


def zoom(w, *keys):
    return [w[k].zoom() for k in keys]


# TODO need to support lists
class Zoomable:
    def __init__(self, parent, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.parent = parent

    # TODO not sure, maybe do it via del??
    # TODO need to make sure they are in proper order? object should be last..
    @property
    def dependants(self):
        raise NotImplementedError

    def ignore(self) -> None:
        self.consume_all()

    def consume_all(self) -> None:
        for d in self.dependants:
            d.consume_all()
        self.consume()

    def consume(self) -> None:
        assert self.parent is not None
        self.parent._remove(self)

    def zoom(self) -> Zoomable:
        self.consume()
        return self

    def _remove(self, xx):
        raise NotImplementedError

    def this_consumed(self):
        raise NotImplementedError


class Wdict(Zoomable, OrderedDict):
    def _remove(self, xx):
        keys = [k for k, v in self.items() if v is xx]
        assert len(keys) == 1
        del self[keys[0]]

    @property
    def dependants(self):
        return list(self.values())

    def this_consumed(self):
        return len(self) == 0

    # TODO specify mypy type for the index special method?


class Wlist(Zoomable, list):
    def _remove(self, xx):
        self.remove(xx)

    @property
    def dependants(self):
        return list(self)

    def this_consumed(self):
        return len(self) == 0


class Wvalue(Zoomable):
    def __init__(self, parent, value: Any) -> None:
        super().__init__(parent)
        self.value = value

    @property
    def dependants(self):
        return []

    def this_consumed(self):
        return True  # TODO not sure..

    def __repr__(self):
        return 'WValue{' + repr(self.value) + '}'


def _wrap(j, parent=None) -> tuple[Zoomable, list[Zoomable]]:
    res: Zoomable
    cc: list[Zoomable]
    if isinstance(j, dict):
        res = Wdict(parent)
        cc = [res]
        for k, v in j.items():
            vv, c = _wrap(v, parent=res)
            res[k] = vv
            cc.extend(c)
        return res, cc
    elif isinstance(j, list):
        res = Wlist(parent)
        cc = [res]
        for i in j:
            ii, c = _wrap(i, parent=res)
            res.append(ii)
            cc.extend(c)
        return res, cc
    elif isinstance(j, (int, float, str, type(None))):
        res = Wvalue(parent, j)
        return res, [res]
    else:
        raise RuntimeError(f'Unexpected type: {type(j)} {j}')


class UnconsumedError(Exception):
    pass


# TODO think about error policy later...
@contextmanager
def wrap(j, *, throw=True) -> Iterator[Zoomable]:
    w, children = _wrap(j)

    yield w

    for c in children:
        if not c.this_consumed():  # TODO hmm. how does it figure out if it's consumed???
            if throw:
                # TODO need to keep a full path or something...
                raise UnconsumedError(f'''
Expected {c} to be fully consumed by the parser.
'''.lstrip())
            else:
                # TODO log?
                pass


def test_unconsumed() -> None:
    import pytest

    with pytest.raises(UnconsumedError):
        with wrap({'a': 1234}) as w:
            w = cast(Wdict, w)

    with pytest.raises(UnconsumedError):
        with wrap({'c': {'d': 2222}}) as w:
            w = cast(Wdict, w)
            _d = w['c']['d'].zoom()


def test_consumed() -> None:
    with wrap({'a': 1234}) as w:
        w = cast(Wdict, w)
        _a = w['a'].zoom()

    with wrap({'c': {'d': 2222}}) as w:
        w = cast(Wdict, w)
        c = w['c'].zoom()
        _d = c['d'].zoom()


def test_types() -> None:
    # (string, number, object, array, boolean or nul
    with wrap({'string': 'string', 'number': 3.14, 'boolean': True, 'null': None, 'list': [1, 2, 3]}) as w:
        w = cast(Wdict, w)
        w['string'].zoom()
        w['number'].consume()
        w['boolean'].zoom()
        w['null'].zoom()
        for x in list(w['list'].zoom()):  # TODO eh. how to avoid the extra list thing?
            x.consume()


def test_consume_all() -> None:
    with wrap({'aaa': {'bbb': {'hi': 123}}}) as w:
        w = cast(Wdict, w)
        aaa = w['aaa'].zoom()
        aaa['bbb'].consume_all()


def test_consume_few() -> None:
    import pytest

    pytest.skip('Will think about it later..')
    with wrap({'important': 123, 'unimportant': 'whatever'}) as w:
        w = cast(Wdict, w)
        w['important'].zoom()
        w.consume_all()
        # TODO hmm, we want smth like this to work..


def test_zoom() -> None:
    import pytest

    with wrap({'aaa': 'whatever'}) as w:
        w = cast(Wdict, w)
        with pytest.raises(KeyError):
            w['nosuchkey'].zoom()
        w['aaa'].zoom()


# TODO type check this...

# TODO feels like the whole thing kind of unnecessarily complex
# - cons:
#     - in most cases this is not even needed? who cares if we miss a few attributes?
# - pro: on the other hand it could be interesting to know about new attributes in data,
#        and without this kind of processing we wouldn't even know
# alternatives
# - manually process data
#   e.g. use asserts, dict.pop and dict.values() methods to unpack things
#   - pros:
#     - very simple, since uses built in syntax
#     - very performant, as fast as it gets
#     - very flexible, easy to adjust behaviour
#   - cons:
#     - can forget to assert about extra entities etc, so error prone
#     - if we do something like =assert j.pop('status') == 200, j=, by the time assert happens we already popped item -- makes error handling harder
#     - a bit verbose.. so probably requires some helper functions though (could be much leaner than current konsume though)
#     - if we assert, then terminates parsing too early, if we're defensive then inflates the code a lot with if statements
#       - TODO perhaps combine warnings somehow or at least only emit once per module?
#       - hmm actually tbh if we carefully go through everything and don't make copies, then only requires one assert at the very end?
#   - TODO this is kinda useful? https://discuss.python.org/t/syntax-for-dictionnary-unpacking-to-variables/18718
#     operator.itemgetter?
#   - TODO can use match operator in python for this? quite nice actually! and allows for dynamic behaviour
#     only from 3.10 tho, and gonna be tricky to do dynamic defensive behaviour with this
#   - TODO in a sense, blenser already would hint if some meaningful fields aren't being processed? only if they are changing though
# - define a "schema" for data, then just recursively match data against the schema?
#   possibly pydantic already does something like that? not sure about performance though
#   pros:
#     - much simpler to extend and understand what's going on
#   cons:
#     - more rigid, so it becomes tricky to do dynamic stuff (e.g. if schema actually changes)



"""
Various tests that are checking behaviour of user config wrt to various things
"""

import os
import sys
from pathlib import Path

import pytest
import pytz

import my.config
from my.core import notnone
from my.demo import items, make_config

from .common import tmp_environ_set

# TODO would be nice to randomize test order here to catch various config issues


# run the same test multiple times to make sure there are not issues with import order etc
@pytest.mark.parametrize('run_id', ['1', '2'])
def test_override_config(tmp_path: Path, run_id: str) -> None:
    class user_config:
        username = f'user_{run_id}'
        data_path = f'{tmp_path}/*.json'

    my.config.demo = user_config  # type: ignore[misc, assignment]

    [item1, item2] = items()
    assert item1.username == f'user_{run_id}'
    assert item2.username == f'user_{run_id}'


@pytest.mark.skip(reason="won't work at the moment because of inheritance")
def test_dynamic_config_simplenamespace(tmp_path: Path) -> None:
    from types import SimpleNamespace

    user_config = SimpleNamespace(
        username='user3',
        data_path=f'{tmp_path}/*.json',
    )
    my.config.demo = user_config  # type: ignore[misc, assignment]

    cfg = make_config()

    assert cfg.username == 'user3'


def test_mixin_attribute_handling(tmp_path: Path) -> None:
    """
    Tests that arbitrary mixin attributes work with our config handling pattern
    """

    nytz = pytz.timezone('America/New_York')

    class user_config:
        # check that override is taken into the account
        timezone = nytz

        irrelevant = 'hello'

        username = 'UUU'
        data_path = f'{tmp_path}/*.json'

    my.config.demo = user_config  # type: ignore[misc, assignment]

    cfg = make_config()

    assert cfg.username == 'UUU'

    # mypy doesn't know about it, but the attribute is there
    assert getattr(cfg, 'irrelevant') == 'hello'

    # check that overridden default attribute is actually getting overridden
    assert cfg.timezone == nytz

    [item1, item2] = items()
    assert item1.username == 'UUU'
    assert notnone(item1.dt.tzinfo).zone == nytz.zone  # type: ignore[attr-defined]
    assert item2.username == 'UUU'
    assert notnone(item2.dt.tzinfo).zone == nytz.zone  # type: ignore[attr-defined]


# use multiple identical tests to make sure there are no issues with cached imports etc
@pytest.mark.parametrize('run_id', ['1', '2'])
def test_dynamic_module_import(tmp_path: Path, run_id: str) -> None:
    """
    Test for dynamic hackery in config properties
     e.g. importing some external modules
    """

    ext = tmp_path / 'external'
    ext.mkdir()
    (ext / '__init__.py').write_text(
        '''
def transform(x):
    from .submodule import do_transform
    return do_transform(x)

'''
    )
    (ext / 'submodule.py').write_text(
        f'''
def do_transform(x):
    return {{"total_{run_id}": sum(x.values())}}
'''
    )

    class user_config:
        username = 'someuser'
        data_path = f'{tmp_path}/*.json'
        external = f'{ext}'

    my.config.demo = user_config  # type: ignore[misc, assignment]

    [item1, item2] = items()
    assert item1.raw == {f'total_{run_id}': 1 + 123}, item1
    assert item2.raw == {f'total_{run_id}': 2 + 456}, item2

    # need to reset these modules, otherwise they get cached
    # kind of relevant to my.core.cfg.tmp_config
    sys.modules.pop('external', None)
    sys.modules.pop('external.submodule', None)


@pytest.mark.parametrize('run_id', ['1', '2'])
def test_my_config_env_variable(tmp_path: Path, run_id: str) -> None:
    """
    Tests handling of MY_CONFIG variable
    """

    # ugh. so by this point, my.config is already loaded (default stub), so we need to unload it
    sys.modules.pop('my.config', None)
    # but my.config itself relies on my.core.init hook, so unless it's reloaded too it wouldn't help
    sys.modules.pop('my.core', None)
    sys.modules.pop('my.core.init', None)
    # it's a bit of a mouthful of course, but in most cases MY_CONFIG would be set once
    #  , and before hpi runs, so hopefully it's not a huge deal
    cfg_dir = tmp_path / 'my'
    cfg_file = cfg_dir / 'config.py'
    cfg_dir.mkdir()

    cfg_file.write_text(
        f'''
# print("IMPORTING CONFIG {run_id}")
class demo:
    username = 'xxx_{run_id}'
    data_path = r'{tmp_path}{os.sep}*.json'  # need raw string for windows...
'''
    )

    with tmp_environ_set('MY_CONFIG', str(tmp_path)):
        [item1, item2] = items()
        assert item1.username == f'xxx_{run_id}'
        assert item2.username == f'xxx_{run_id}'

        # sigh.. so this is cached in sys.path
        #  so it takes precedence later during next import, not giving the MY_CONFIG hook
        #  (imported from builtin my.config) to kick in
        sys.path.remove(str(tmp_path))

        # FIXME ideally this shouldn't be necessary?
        #  remove this after we fixup my.tests.reddit and my.tests.commits
        #  (they were failing ci when running all tests)
        sys.modules.pop('my.config', None)


@pytest.fixture(autouse=True)
def prepare_data(tmp_path: Path):
    (tmp_path / 'data.json').write_text(
        '''
[
    {"key": 1, "value": 123},
    {"key": 2, "value": 456}
]
'''
    )



"""
Helper to test various timezone/location dependent things
"""

from datetime import date, datetime, timezone
from pathlib import Path

import pytest
from more_itertools import one

from my.core.cfg import tmp_config


@pytest.fixture(autouse=True)
def config(tmp_path: Path):
    # TODO could just pick a part of shared config? not sure
    _takeout_path = _prepare_takeouts_dir(tmp_path)

    class google:
        takeout_path = _takeout_path

    class location:
        # fmt: off
        home = (
            # supports ISO strings
            ('2005-12-04'                                       , (42.697842, 23.325973)), # Bulgaria, Sofia
            # supports date/datetime objects
            (date(year=1980, month=2, day=15)                   , (40.7128  , -74.0060 )), # NY
            # check tz handling..
            (datetime.fromtimestamp(1600000000, tz=timezone.utc), (55.7558  , 37.6173  )), # Moscow, Russia
        )
        # fmt: on
        # note: order doesn't matter, will be sorted in the data provider

    class time:
        class tz:
            class via_location:
                fast = True  # some tests rely on it

    with tmp_config() as cfg:
        cfg.google = google
        cfg.location = location
        cfg.time = time
        yield cfg


def _prepare_takeouts_dir(tmp_path: Path) -> Path:
    from .common import testdata

    testdata_dir = testdata()
    try:
        track = one(testdata_dir.rglob('italy-slovenia-2017-07-29.json'))
    except ValueError as e:
        raise RuntimeError(f'testdata not found in {testdata_dir}, setup git submodules?') from e

    # todo ugh. unnecessary zipping, but at the moment takeout provider doesn't support plain dirs
    import zipfile

    with zipfile.ZipFile(tmp_path / 'takeout.zip', 'w') as zf:
        zf.writestr('Takeout/Location History/Location History.json', track.read_bytes())
    return tmp_path



'''
Last.fm scrobbles
'''

from abc import abstractmethod
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Protocol

from my.core import Json, Paths, Stats, datetime_aware, get_files, make_logger, stat
from my.core.cachew import mcachew
from my.core.json import json_loads

logger = make_logger(__name__)


class config(Protocol):
    @property
    @abstractmethod
    def export_path(self) -> Paths:
        """
        Uses [[https://github.com/karlicoss/lastfm-backup][lastfm-backup]] outputs
        """
        raise NotImplementedError


def make_config() -> config:
    from my.config import lastfm as user_config

    class combined_config(user_config, config): ...

    return combined_config()


def inputs() -> Sequence[Path]:
    cfg = make_config()
    return get_files(cfg.export_path)


# TODO memoised properties?
# TODO lazy mode and eager mode?
# lazy is a bit nicer in terms of more flexibility and less processing?
# eager is a bit more explicit for error handling


@dataclass
class Scrobble:
    raw: Json

    @property
    def dt(self) -> datetime_aware:
        ts = int(self.raw['date'])
        return datetime.fromtimestamp(ts, tz=timezone.utc)

    @property
    def artist(self) -> str:
        return self.raw['artist']

    @property
    def name(self) -> str:
        return self.raw['name']

    @property
    def track(self) -> str:
        return f'{self.artist} — {self.name}'

    # TODO __repr__, __str__
    # TODO could also be nice to make generic? maybe even depending on eagerness


@mcachew(depends_on=inputs)
def scrobbles() -> Iterable[Scrobble]:
    last = max(inputs())
    logger.info(f'loading data from {last}')
    j = json_loads(last.read_bytes())

    for raw in reversed(j):
        yield Scrobble(raw=raw)


def stats() -> Stats:
    return stat(scrobbles)


def fill_influxdb() -> None:
    from my.core import influxdb

    # todo needs to be more automatic
    sd = (
        {
            'dt': x.dt,
            'track': x.track,
        }
        for x in scrobbles()
    )
    influxdb.fill(sd, measurement=__name__)



from my.tests.common import skip_if_not_karlicoss as pytestmark  # isort: skip


def test() -> None:
    from my.hypothesis import highlights, pages
    assert len(list(pages())) > 10
    assert len(list(highlights())) > 10



"""
Tests for LEGACY location provider

Keeping for now for backwards compatibility
"""

from pathlib import Path

import pytest
from more_itertools import one

from my.core.cfg import tmp_config
from my.location.google import locations


def test_google_locations() -> None:
    locs = list(locations())
    assert len(locs) == 3810, len(locs)

    last = locs[-1]
    assert last.dt.strftime('%Y%m%d %H:%M:%S') == '20170802 13:01:56'  # should be utc
    # todo approx
    assert last.lat == 46.5515350
    assert last.lon == 16.4742742
    # todo check altitude


@pytest.fixture(autouse=True)
def prepare(tmp_path: Path):

    # TODO could just pick a part of shared config? not sure
    _takeout_path = _prepare_takeouts_dir(tmp_path)

    class google:
        takeout_path = _takeout_path

    with tmp_config() as config:
        config.google = google
        yield


def _prepare_takeouts_dir(tmp_path: Path) -> Path:
    from ..common import testdata

    try:
        track = one(testdata().rglob('italy-slovenia-2017-07-29.json'))
    except ValueError as e:
        raise RuntimeError('testdata not found, setup git submodules?') from e

    # todo ugh. unnecessary zipping, but at the moment takeout provider doesn't support plain dirs
    import zipfile

    with zipfile.ZipFile(tmp_path / 'takeout.zip', 'w') as zf:
        zf.writestr('Takeout/Location History/Location History.json', track.read_bytes())
    return tmp_path



"""
Utils specific to hpi core, shouldn't really be used by HPI modules
"""
from __future__ import annotations


def _is_editable(package_name: str) -> bool:
    import importlib.metadata

    dist = importlib.metadata.distribution(package_name)
    dist_files = dist.files or []
    for path in dist_files:
        if str(path).endswith('.pth'):
            return True
    return False


def warn_if_not_using_src_layout(path: list[str]) -> None:
    contains_src = any('/src/my/' in p for p in path)
    if contains_src:
        return

    # __package__ won't work because it's gonna be 'my' rather than 'hpi'
    # seems like it's quite annoying to get distribition name from within the package, so easier to hardcode..
    distribution_name = 'hpi'
    try:
        editable = _is_editable(distribution_name)
    except:
        # it would be annoying if this somehow fails during the very first HPI import...
        # so just make defensive
        return

    if not editable:
        # nothing to check
        return


    from . import warnings

    MSG = '''
    Seems that your HPI is installed as editable and uses flat layout ( my/ directory next to .git folder).
    This was the case in older HPI versions (pre-20250123), but now src-layout is recommended ( src/my/ directory next to .git folder).

    Reinstall your HPI as editable again via 'pip install --editable /path/to/hpi'.
    See https://github.com/karlicoss/HPI/issues/316 for more info.
    '''

    warnings.high(MSG)



"""
Parses browser history using [[http://github.com/purarue/browserexport][browserexport]]
"""

REQUIRES = ["browserexport"]

from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from pathlib import Path

from browserexport.merge import Visit, read_and_merge

from my.core import (
    Paths,
    Stats,
    get_files,
    make_logger,
    stat,
)
from my.core.cachew import mcachew

from .common import _patch_browserexport_logs

import my.config  # isort: skip


@dataclass
class config(my.config.browser.export):
    # path[s]/glob to your backed up browser history sqlite files
    export_path: Paths


logger = make_logger(__name__)
_patch_browserexport_logs(logger.level)


# all of my backed up databases
def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


@mcachew(depends_on=inputs, logger=logger)
def history() -> Iterator[Visit]:
    yield from read_and_merge(inputs())


def stats() -> Stats:
    return {**stat(history)}



"""
Twitter data (tweets and favorites). Uses [[https://github.com/twintproject/twint][Twint]] data export.
"""
from collections.abc import Iterator
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import NamedTuple

from my.core import Json, LazyLogger, Paths, Res, Stats, datetime_aware, get_files, stat
from my.core.cfg import make_config
from my.core.sqlite import sqlite_connection

from my.config import twint as user_config  # isort: skip

# TODO move to twitter.twint config structure

@dataclass
class twint(user_config):
    export_path: Paths # path[s]/glob to the twint Sqlite database

####

config = make_config(twint)


log = LazyLogger(__name__)


def get_db_path() -> Path:
    return max(get_files(config.export_path))


from .common import TweetId, permalink


class Tweet(NamedTuple):
    row: Json

    @property
    def id_str(self) -> TweetId:
        return self.row['id_str']

    @property
    def created_at(self) -> datetime_aware:
        seconds = self.row['created_at'] / 1000
        tz = timezone.utc
        # NOTE: UTC seems to be the case at least for the older version of schema I was using
        # in twint, it was extracted from "data-time-ms" field in the scraped HML
        # https://github.com/twintproject/twint/blob/e3345426eb24154ff084be22e4fed5cfa4631930/twint/tweet.py#L85
        #
        # I checked against twitter archive which is definitely UTC, and it seems to match
        # also seems that other people are treating it as utc, e.g.
        # https://github.com/thomasancheriyil/Red-Tide-Detection-based-on-Twitter/blob/beb200be60cc66dcbc394e670513715509837812/python/twitterGapParse.py#L61-L62
        #
        # twint is also saving 'timezone', but this is local machine timezone at the time of scraping?
        # perhaps they thought date-time-ms was local time... or just kept it just in case (they are keeping lots on unnecessary stuff in the db)
        return datetime.fromtimestamp(seconds, tz=tz)

    @property
    def screen_name(self) -> str:
        return self.row['screen_name']

    @property
    def text(self) -> str:
        text = self.row['tweet']
        mentions_s = self.row['mentions']
        if len(mentions_s) > 0:
            # at some point for no apparent reasions mentions stopped appearing from tweet text in twint
            # note that the order is still inconsisnent against twitter archive, but not much we can do
            mentions = mentions_s.split(',')
            for m in mentions:
                # ugh. sometimes they appear as lowercase in text, sometimes not..
                if m.lower() not in text.lower():
                    text = f'@{m} ' + text
        return text

    @property
    def urls(self) -> list[str]:
        ustr = self.row['urls']
        if len(ustr) == 0:
            return []
        return ustr.split(',')

    @property
    def permalink(self) -> str:
        return permalink(screen_name=self.screen_name, id=self.id_str)


    # TODO urls
    def __repr__(self):
        return f'Tweet(id_str={self.id_str}, created_at={self.created_at}, text={self.text})'

# https://github.com/twintproject/twint/issues/196
# ugh. so it dumps everything in tweet table, and there is no good way to tell between fav/original tweet.
# it might result in some tweets missing from the timeline if you happened to like them...
# not sure what to do with it
# alternatively, could ask the user to run separate databases for tweets and favs?
# TODO think about it

_QUERY = '''
SELECT T.*
FROM      tweets    as T
LEFT JOIN favorites as F
ON    T.id_str = F.tweet_id
WHERE {where}
ORDER BY T.created_at
'''


def tweets() -> Iterator[Res[Tweet]]:
    with sqlite_connection(get_db_path(), immutable=True, row_factory='dict') as db:
        res = db.execute(_QUERY.format(where='F.tweet_id IS NULL'))
        yield from map(Tweet, res)


def likes() -> Iterator[Res[Tweet]]:
    with sqlite_connection(get_db_path(), immutable=True, row_factory='dict') as db:
        res = db.execute(_QUERY.format(where='F.tweet_id IS NOT NULL'))
        yield from map(Tweet, res)


def stats() -> Stats:
    return {
        **stat(tweets),
        **stat(likes),
    }



'''
Just a demo module for testing and documentation purposes
'''
from collections.abc import Iterator
from dataclasses import dataclass

from my.config import simple as user_config
from my.core import make_config


@dataclass
class simple(user_config):
    count: int


config = make_config(simple)


def items() -> Iterator[int]:
    yield from range(config.count)



from __future__ import annotations

from concurrent.futures import Executor, Future
from typing import Any, Callable, TypeVar

from ..compat import ParamSpec

_P = ParamSpec('_P')
_T = TypeVar('_T')


# https://stackoverflow.com/a/10436851/706389
class DummyExecutor(Executor):
    """
    This is useful if you're already using Executor for parallelising,
     but also want to provide an option to run the code serially (e.g. for debugging)
    """

    def __init__(self, max_workers: int | None = 1) -> None:
        self._shutdown = False
        self._max_workers = max_workers

    def submit(self, fn: Callable[_P, _T], /, *args: _P.args, **kwargs: _P.kwargs) -> Future[_T]:
        if self._shutdown:
            raise RuntimeError('cannot schedule new futures after shutdown')

        f: Future[Any] = Future()
        try:
            result = fn(*args, **kwargs)
        except KeyboardInterrupt:
            raise
        except BaseException as e:
            f.set_exception(e)
        else:
            f.set_result(result)

        return f

    def shutdown(self, wait: bool = True, **kwargs) -> None:  # noqa: FBT001,FBT002,ARG002
        self._shutdown = True



"""
Whatsapp data from Android app database (in =/data/data/com.whatsapp/databases/msgstore.db=)
"""

from __future__ import annotations

import sqlite3
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Union

from my.core import Paths, Res, datetime_aware, get_files, make_config, make_logger
from my.core.common import unique_everseen
from my.core.compat import add_note
from my.core.error import notnone
from my.core.sqlite import sqlite_connection

import my.config  # isort: skip

logger = make_logger(__name__)


@dataclass
class Config(my.config.whatsapp.android):
    # paths[s]/glob to the exported sqlite databases
    export_path: Paths
    my_user_id: str | None = None


config = make_config(Config)


def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


@dataclass(unsafe_hash=True)
class Chat:
    id: str
    # todo not sure how to support renames?
    # could change Chat object itself, but this won't work well with incremental processing..
    name: str | None


@dataclass(unsafe_hash=True)
class Sender:
    id: str
    name: str | None


@dataclass(unsafe_hash=True)
class Message:
    chat: Chat
    id: str
    dt: datetime_aware
    sender: Sender
    text: str | None


Entity = Union[Chat, Sender, Message]


def _process_db(db: sqlite3.Connection) -> Iterator[Entity]:
    # TODO later, split out Chat/Sender objects separately to safe on object creation, similar to other android data sources

    try:
        db.execute('SELECT jid_row_id FROM chat_view')
    except sqlite3.OperationalError as oe:
        if 'jid_row_id' not in str(oe):
            raise oe
        new_version_202410 = False
    else:
        new_version_202410 = True

    if new_version_202410:
        chat_id_col = 'jid.raw_string'
        jid_join = 'JOIN jid ON jid._id == chat_view.jid_row_id'
    else:
        chat_id_col = 'chat_view.raw_string_jid'
        jid_join = ''

    chats = {}
    for r in db.execute(
        f'''
    SELECT {chat_id_col} AS chat_id, subject
    FROM chat_view {jid_join}
    WHERE chat_id IS NOT NULL /* seems that it might be null for chats that are 'recycled' (the db is more like an LRU cache) */
    '''
    ):
        chat_id = r['chat_id']
        subject = r['subject']
        chat = Chat(
            id=chat_id,
            name=subject,
        )
        yield chat
        chats[chat.id] = chat

    senders = {}
    for r in db.execute(
        '''
    SELECT _id, raw_string
    FROM jid
    '''
    ):
        # TODO seems that msgstore.db doesn't have contact names
        # perhaps should extract from wa.db and match against wa_contacts.jid?
        # TODO these can also be chats? not sure if need to include...
        s = Sender(
            id=r['raw_string'],
            name=None,
        )
        yield s
        senders[r['_id']] = s

    # NOTE: hmm, seems that message_view or available_message_view use lots of NULL as ...
    # so even if it seems as if it has a column (e.g. for attachment path), there is actually no such data
    # so makes more sense to just query message column directly
    for r in db.execute(
        f'''
    SELECT
        {chat_id_col} AS chat_id,
        M.key_id, M.timestamp,
        sender_jid_row_id,
        M.from_me,
        M.text_data,
        MM.file_path,
        MM.file_size,
        M.message_type
    FROM      message       AS M
    LEFT JOIN chat_view     ON M.chat_row_id = chat_view._id
    {jid_join}
    left JOIN message_media AS MM ON M._id = MM.message_row_id
    WHERE M.key_id != -1 /*  key_id -1 is some sort of fake message where everything is null */
           /* type 7 seems to be some dummy system message.
              sometimes contain chat name, but usually null, so ignore them
              for normal messages it's 0
           */
          AND M.message_type != 7
    ORDER BY M.timestamp
    '''
    ):
        msg_id: str = notnone(r['key_id'])
        ts: int = notnone(r['timestamp'])
        dt = datetime.fromtimestamp(ts / 1000, tz=timezone.utc)

        text: str | None = r['text_data']
        media_file_path: str | None = r['file_path']
        media_file_size: int | None = r['file_size']

        message_type = r['message_type']

        if text is None:
            # fmt: off
            text = {
                5 : '[MAP LOCATION]',
                10: '[MISSED VOICE CALL]',
                15: '[DELETED]',
                16: '[LIVE LOCATION]',
                64: '[DELETED]',  # seems like 'deleted by admin'?
            }.get(message_type)
            # fmt: on

        # check against known msg types
        # fmt: off
        if text is None and message_type not in {
            0,  # normal
            1,  # image
            2,  # voice note
            3,  # video
            7,  # "system" message, e.g. chat name
            8,  # document
            9,  # also document?
            13, # animated gif?
            20, # webp/sticker?
        }:
            text = f"[UNKNOWN TYPE {message_type}]"
        # fmt: on

        if media_file_size is not None:
            # this is always not null for message_media table
            # however media_file_path sometimes may be none
            mm = f'MEDIA: {media_file_path}'
            if text is None:
                text = mm
            else:
                text = text + '\n' + mm

        from_me = r['from_me'] == 1

        chat_id = r['chat_id']
        if chat_id is None:
            # ugh, I think these might have been edited messages? unclear..
            logger.warning(f"CHAT ID IS NONE, WTF?? {dt} {ts} {text}")
            continue
        chat = chats[chat_id]

        sender_row_id = r['sender_jid_row_id']
        if sender_row_id == 0:
            # seems that it's always 0 for 1-1 chats
            # for group chats our own id is still 0, but other ids are properly set
            if from_me:
                myself_user_id = config.my_user_id or 'MYSELF_USER_ID'
                sender = Sender(id=myself_user_id, name=None)  # TODO set my own name as well?
            else:
                sender = Sender(id=chat.id, name=None)
        else:
            sender = senders[sender_row_id]

        m = Message(chat=chat, id=msg_id, dt=dt, sender=sender, text=text)
        yield m


def _entities() -> Iterator[Res[Entity]]:
    paths = inputs()
    total = len(paths)
    width = len(str(total))
    for idx, path in enumerate(paths):
        logger.info(f'processing [{idx:>{width}}/{total:>{width}}] {path}')
        with sqlite_connection(path, immutable=True, row_factory='row') as db:
            try:
                yield from _process_db(db)
            except Exception as e:
                add_note(e, f'^ while processing {path}')


def entities() -> Iterator[Res[Entity]]:
    return unique_everseen(_entities)


def messages() -> Iterator[Res[Message]]:
    # TODO hmm, specify key=lambda m: m.id?
    # not sure since might be useful to keep track of sender changes etc
    # probably best not to, or maybe query messages/senders separately and merge later?
    for e in entities():
        if isinstance(e, (Exception, Message)):
            yield e



"""
Twitter data from Talon app database (in =/data/data/com.klinker.android.twitter_l/databases/=)
"""

from __future__ import annotations

import re
import sqlite3
from abc import abstractmethod
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Union

from my.core import Paths, Res, datetime_aware, get_files
from my.core.common import unique_everseen
from my.core.sqlite import sqlite_connection

from .common import TweetId, permalink


class config:
    @property
    @abstractmethod
    def export_path(self) -> Paths:
        raise NotImplementedError


def make_config() -> config:
    from my.config import twitter as user_config

    class combined_config(user_config.talon, config):
        pass

    return combined_config()


def inputs() -> Sequence[Path]:
    return get_files(make_config().export_path)


@dataclass(unsafe_hash=True)
class Tweet:
    id_str: TweetId
    created_at: datetime_aware
    screen_name: str
    text: str
    urls: Sequence[str]

    @property
    def permalink(self) -> str:
        return permalink(screen_name=self.screen_name, id=self.id_str)


# meh... just wrappers to tell apart tweets from favorites...
@dataclass(unsafe_hash=True)
class _IsTweet:
    tweet: Tweet


@dataclass(unsafe_hash=True)
class _IsFavorire:
    tweet: Tweet


Entity = Union[_IsTweet, _IsFavorire]


def _entities() -> Iterator[Res[Entity]]:
    for f in inputs():
        yield from _process_one(f)


def _process_one(f: Path) -> Iterator[Res[Entity]]:
    handlers = {
        'user_tweets.db': _process_user_tweets,
        'favorite_tweets.db': _process_favorite_tweets,
    }
    fname = f.name
    handler = handlers.get(fname)
    if handler is None:
        yield RuntimeError(f"Could not find handler for {fname}")
        return
    with sqlite_connection(f, immutable=True, row_factory='row') as db:
        yield from handler(db)


def _process_user_tweets(db: sqlite3.Connection) -> Iterator[Res[Entity]]:
    # dunno why it's called 'lists'
    for r in db.execute('SELECT * FROM lists ORDER BY time'):
        try:
            yield _IsTweet(_parse_tweet(r))
        except Exception as e:
            yield e


def _process_favorite_tweets(db: sqlite3.Connection) -> Iterator[Res[Entity]]:
    for r in db.execute('SELECT * FROM favorite_tweets ORDER BY time'):
        try:
            yield _IsFavorire(_parse_tweet(r))
        except Exception as e:
            yield e


def _parse_tweet(row: sqlite3.Row) -> Tweet:
    # ok so looks like it's tz aware..
    # https://github.com/klinker24/talon-for-twitter-android/blob/c3b0612717ba3ea93c0cae6d907d7d86d640069e/app/src/main/java/com/klinker/android/twitter_l/data/sq_lite/FavoriteTweetsDataSource.java#L95
    # uses https://docs.oracle.com/javase/7/docs/api/java/util/Date.html#getTime()
    # and it's created here, so looks like it's properly parsed from the api
    # https://github.com/Twitter4J/Twitter4J/blob/8376fade8d557896bb9319fb46e39a55b134b166/twitter4j-core/src/internal-json/java/twitter4j/ParseUtil.java#L69-L79
    created_at = datetime.fromtimestamp(row['time'] / 1000, tz=timezone.utc)
    text = row['text']

    # try explanding URLs.. sadly there are no positions in the db
    urls = row['other_url'].split()
    if len(urls) > 0:
        ellipsis = '...'
        # might have something collapsed
        # e.g. deepmind.com/blog/article/Comp...
        # NOTE: need a one character of lookahead to split on ellipsis.. hence ?=
        for short in re.findall(r'(?:^|\s)([\S]+)' + re.escape(ellipsis) + r'(?=\s|$)', text):
            for full in urls:
                if short in full:
                    text = text.replace(short + ellipsis, full)
                    break
    #

    screen_name = row['screen_name']
    # considering id_str is referring to the retweeter's tweet (rather than the original tweet)
    # makes sense for the permalink to contain the retweeter as well
    # also makes it more compatible to twitter archive
    # a bit sad to lose structured information about RT, but then again we could always just parse it..
    retweeter = row['retweeter']
    if len(retweeter) > 0:
        text = f'RT @{screen_name}: {text}'
        screen_name = retweeter

    return Tweet(
        id_str=str(row['tweet_id']),
        created_at=created_at,
        screen_name=screen_name,
        text=text,
        # todo hmm text sometimes is trimmed with ellipsis? at least urls
        urls=tuple(u for u in row['other_url'].split(' ') if len(u.strip()) > 0),
    )


def tweets() -> Iterator[Res[Tweet]]:
    for x in unique_everseen(_entities):
        if isinstance(x, Exception):
            yield x
        elif isinstance(x, _IsTweet):
            yield x.tweet


def likes() -> Iterator[Res[Tweet]]:
    for x in unique_everseen(_entities):
        if isinstance(x, Exception):
            yield x
        elif isinstance(x, _IsFavorire):
            yield x.tweet


# TODO maybe should combine all public iterators into a stats()



"""
Github data (uses [[https://github.com/settings/admin][official GDPR export]])
"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Iterator, Sequence
from pathlib import Path
from typing import Any

from my.core import Paths, Res, Stats, get_files, make_logger, stat, warnings
from my.core.compat import add_note
from my.core.json import json_loads

from .common import Event, EventIds, parse_dt

logger = make_logger(__name__)


class config:
    @property
    @abstractmethod
    def gdpr_dir(self) -> Paths:
        raise NotImplementedError


def make_config() -> config:
    # TODO later, use a separate user config? (github_gdpr)
    from my.config import github as user_config

    class combined_config(user_config, config):
        pass

    return combined_config()


def inputs() -> Sequence[Path]:
    gdpr_dir = make_config().gdpr_dir
    res = get_files(gdpr_dir)
    schema_json = [f for f in res if f.name == 'schema.json']
    was_unpacked = len(schema_json) > 0
    if was_unpacked:
        # 'legacy' behaviour, we've been passed an extracted export directory
        # although in principle nothing wrong with running against a directory with several unpacked archives
        # so need to think how to support that in the future as well
        return [schema_json[0].parent]
    # otherwise, should contain a bunch of archives?
    # not sure if need to warn if any of them aren't .tar.gz?
    return res


def events() -> Iterator[Res[Event]]:
    last = max(inputs())

    logger.info(f'extracting data from {last}')

    root: Path | None = None

    if last.is_dir():  # if it's already CPath, this will match it
        root = last
    else:
        try:
            from kompress import CPath

            root = CPath(last)
            assert len(list(root.iterdir())) > 0  # trigger to check if we have the kompress version with targz support
        except Exception as e:
            logger.exception(e)
            warnings.high("Upgrade 'kompress' to latest version with native .tar.gz support. Falling back to unpacking to tmp dir.")

    if root is None:
        from my.core.structure import match_structure

        with match_structure(last, expected=()) as res:  # expected=() matches it regardless any patterns
            [root] = res
            yield from _process_one(root)
    else:
        yield from _process_one(root)


def _process_one(root: Path) -> Iterator[Res[Event]]:
    files = sorted(root.glob('*.json'))  # looks like all files are in the root

    # fmt: off
    handler_map = {
        'schema'       : None,
        'issue_events_': None,  # eh, doesn't seem to have any useful bodies
        'attachments_' : None,  # not sure if useful
        'users'        : None,  # just contains random users
        'bots'         : None,  # just contains random bots
        'repositories_'  : _parse_repository,
        'issue_comments_': _parse_issue_comment,
        'issues_'        : _parse_issue,
        'pull_requests_' : _parse_pull_request,
        'projects_'      : _parse_project,
        'releases_'      : _parse_release,
        'commit_comments': _parse_commit_comment,
        ## TODO need to handle these
        'pull_request_review_comments_': None,
        'pull_request_review_threads_': None,
        'pull_request_reviews_': None,
        ##
        'repository_files_': None,  # repository artifacts, probs not very useful
        'discussion_categories_': None,  # doesn't seem to contain any useful info, just some repo metadata
        'organizations_': None,  # no useful info, just some org metadata
    }
    # fmt: on
    for f in files:
        logger.info(f'{f} : processing...')
        handler: Any
        for prefix, h in handler_map.items():
            if not f.name.startswith(prefix):
                continue
            handler = h
            break
        else:
            yield RuntimeError(f'Unhandled file: {f}')
            continue

        if handler is None:
            # ignored
            continue

        j = json_loads(f.read_bytes())
        for r in j:
            try:
                yield handler(r)
            except Exception as e:
                add_note(e, f'^ while processing {f}')
                yield e


def stats() -> Stats:
    return {
        **stat(events),
    }


# TODO typing.TypedDict could be handy here..
def _parse_common(d: dict) -> dict:
    url = d['url']
    body = d.get('body')
    return {
        'dt': parse_dt(d['created_at']),
        'link': url,
        'body': body,
    }


def _parse_repository(d: dict) -> Event:
    pref = 'https://github.com/'
    url = d['url']
    dts = d['created_at']
    rt = d['type']
    assert url.startswith(pref)
    name = url[len(pref) :]
    eid = EventIds.repo_created(dts=dts, name=name, ref_type=rt, ref=None)
    return Event(
        **_parse_common(d),
        summary='created ' + name,
        eid=eid,
    )


# user may be None if the user was deleted
def _is_bot(user: str | None) -> bool:
    if user is None:
        return False
    return "[bot]" in user


def _parse_issue_comment(d: dict) -> Event:
    url = d['url']
    return Event(
        **_parse_common(d),
        summary=f'commented on issue {url}',
        eid='issue_comment_' + url,
        is_bot=_is_bot(d['user']),
    )


def _parse_issue(d: dict) -> Event:
    url = d['url']
    title = d['title']
    return Event(
        **_parse_common(d),
        summary=f'opened issue {title}',
        eid='issue_comment_' + url,
        is_bot=_is_bot(d['user']),
    )


def _parse_pull_request(d: dict) -> Event:
    dts = d['created_at']
    url = d['url']
    title = d['title']
    return Event(
        **_parse_common(d),
        # TODO distinguish incoming/outgoing?
        # TODO action? opened/closed??
        summary=f'opened PR {title}',
        eid=EventIds.pr(dts=dts, action='opened', url=url),
        is_bot=_is_bot(d['user']),
    )


def _parse_project(d: dict) -> Event:
    url = d['url']
    title = d['name']
    is_bot = "[bot]" in d["creator"]
    # TODO: use columns somehow?
    # Doesn't fit with Event schema,
    # is a list of each of the boards
    return Event(
        **_parse_common(d),
        summary=f'created project {title}',
        eid='project_' + url,
        is_bot=is_bot,
    )


def _parse_release(d: dict) -> Event:
    tag = d['tag_name']
    return Event(
        **_parse_common(d),
        summary=f'released {tag}',
        eid='release_' + tag,
    )


def _parse_commit_comment(d: dict) -> Event:
    url = d['url']
    return Event(
        **_parse_common(d),
        summary=f'commented on {url}',
        eid='commit_comment_' + url,
    )



from __future__ import annotations

from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from my.core import Res, Stats, datetime_aware, make_logger, stat, warnings
from my.core.compat import deprecated

logger = make_logger(__name__)


@dataclass
class Watched:
    url: str
    title: str
    when: datetime_aware

    @property
    def eid(self) -> str:
        return f'{self.url}-{self.when.isoformat()}'

    def is_deleted(self) -> bool:
        return self.title == self.url


# todo define error policy?
# although it has one from google takeout module.. so not sure


def watched() -> Iterator[Res[Watched]]:
    emitted: dict[Any, Watched] = {}
    for w in _watched():
        if isinstance(w, Exception):
            yield w  # TODO also make unique?
            continue

        # older exports (e.g. html) didn't have microseconds
        # whereas newer json ones do have them
        # seconds resolution is enough to distinguish watched videos
        # also we're processing takeouts in HPI in reverse order, so first seen watch would contain microseconds, resulting in better data
        without_microsecond = w.when.replace(microsecond=0)

        key = w.url, without_microsecond
        prev = emitted.get(key, None)
        if prev is not None:
            # NOTE: some video titles start with 'Liked ' for liked videos activity
            # but they'd have different timestamp, so fine not to handle them as a special case here
            if w.title in prev.title:
                # often more stuff added to the title, like 'Official Video'
                # in this case not worth emitting the change
                # also handles the case when titles match
                continue
            # otherwise if title changed completely, just emit the change... not sure what else we could do?
            # could merge titles in the 'titles' field and update dynamically? but a bit complicated, maybe later..

            # TODO would also be nice to handle is_deleted here somehow...
            # but for that would need to process data in direct order vs reversed..
            # not sure, maybe this could use a special mode or something?

        emitted[key] = w
        yield w


def _watched() -> Iterator[Res[Watched]]:
    try:
        from google_takeout_parser.models import Activity

        from ..google.takeout.parser import events
    except ModuleNotFoundError as ex:
        logger.exception(ex)
        warnings.high("Please set up my.google.takeout.parser module for better youtube support. Falling back to legacy implementation.")
        yield from _watched_legacy()  # type: ignore[name-defined]
        return

    YOUTUBE_VIDEO_LINK = '://www.youtube.com/watch?v='

    # TODO would be nice to filter, e.g. it's kinda pointless to process Location events
    for e in events():
        if isinstance(e, Exception):
            yield e
            continue

        if not isinstance(e, Activity):
            continue

        url = e.titleUrl

        if url is None:
            continue

        header = e.header

        if header in {'Image Search', 'Search', 'Chrome'}:
            # sometimes results in youtube links.. but definitely not watch history
            continue

        if header not in {'YouTube', 'youtube.com'}:
            # TODO hmm -- wonder if these would end up in dupes in takeout? would be nice to check
            # perhaps this would be easier once we have universal ids
            if YOUTUBE_VIDEO_LINK in url:
                # TODO maybe log in this case or something?
                pass
            continue

        title = e.title

        if header == 'youtube.com' and title.startswith('Visited '):
            continue

        if title.startswith('Searched for') and url.startswith('https://www.youtube.com/results'):
            # search activity, don't need it here
            continue

        if title.startswith('Subscribed to') and url.startswith('https://www.youtube.com/channel/'):
            # todo might be interesting to process somewhere?
            continue

        # all titles contain it, so pointless to include 'Watched '
        # also compatible with legacy titles
        title = title.removeprefix('Watched ')

        # watches originating from some activity end with this, remove it for consistency
        title = title.removesuffix(' - YouTube')

        if YOUTUBE_VIDEO_LINK not in url:
            if 'youtube.com/post/' in url:
                # some sort of channel updates?
                continue
            if 'youtube.com/playlist' in url:
                # 'saved playlist' actions
                continue
            if 'music.youtube.com' in url:
                # todo maybe allow it?
                continue
            if any('From Google Ads' in d for d in e.details):
                # weird, sometimes results in odd urls
                continue

            if title == 'Used YouTube':
                continue

            yield RuntimeError(f'Unexpected url: {e}')
            continue

        # TODO contribute to takeout parser? seems that these still might happen in json data
        title = title.replace("\xa0", " ")

        yield Watched(
            url=url,
            title=title,
            when=e.time,
        )


def stats() -> Stats:
    return stat(watched)


### deprecated stuff (keep in my.media.youtube)

if not TYPE_CHECKING:

    @deprecated("use 'watched' instead")
    def get_watched(*args, **kwargs):
        return watched(*args, **kwargs)

    def _watched_legacy() -> Iterable[Watched]:
        from ..google.takeout.html import read_html
        from ..google.takeout.paths import get_last_takeout

        # todo looks like this one doesn't have retention? so enough to use the last
        path = 'Takeout/My Activity/YouTube/MyActivity.html'
        last = get_last_takeout(path=path)
        if last is None:
            return []

        watches: list[Watched] = []
        for dt, url, title in read_html(last, path):
            watches.append(Watched(url=url, title=title, when=dt))

        # todo hmm they already come sorted.. wonder if should just rely on it..
        return sorted(watches, key=lambda e: e.when)



'''
My cross trainer exercise data, arbitrated from different sources (mainly, Endomondo and manual text notes)

This is probably too specific to my needs, so later I will move it away to a personal 'layer'.
For now it's worth keeping it here as an example and perhaps utility functions might be useful for other HPI modules.
'''

from __future__ import annotations

from datetime import datetime, timedelta

import pytz

from my.config import exercise as config

from ...core.orgmode import Table, TypedTable, collect, parse_org_datetime
from ...core.pandas import DataFrameT
from ...core.pandas import check_dataframe as cdf

# FIXME how to attach it properly?
tz = pytz.timezone('Europe/London')

def tzify(d: datetime) -> datetime:
    assert d.tzinfo is None, d
    return tz.localize(d)


# todo predataframe?? entries??
def cross_trainer_data():
    # FIXME some manual entries in python
    # I guess just convert them to org
    import orgparse
    # todo should use all org notes and just query from them?
    wlog = orgparse.load(config.workout_log)

    [table] = collect(
        wlog,
        lambda n: [] if n.heading != 'Cross training' else [x for x in n.body_rich if isinstance(x, Table)]
    )
    cross_table = TypedTable(table)

    def maybe(f):
        def parse(s):
            if len(s) == 0:
                return None
            return f(s)
        return parse

    def parse_mm_ss(x: str) -> timedelta:
        hs, ms = x.split(':')
        return timedelta(seconds=int(hs) * 60 + int(ms))

    # todo eh. not sure if there is a way of getting around writing code...
    # I guess would be nice to have a means of specifying type in the column? maybe multirow column names??
    # need to look up org-mode standard..
    mappers = {
        'duration': lambda s: parse_mm_ss(s),
        'date'    : lambda s: tzify(parse_org_datetime(s)),
        'comment' : str,
    }
    for row in cross_table.as_dicts:
        # todo make more defensive, fallback on nan for individual fields??
        try:
            d = {}
            for k, v in row.items():
                # todo have something smarter... e.g. allow pandas to infer the type??
                mapper = mappers.get(k, maybe(float))
                d[k] = mapper(v) # type: ignore[operator]
            yield d
        except Exception as e:
            # todo add parsing context
            yield {'error': str(e)}

    # todo hmm, converting an org table directly to pandas kinda makes sense?
    # could have a '.dataframe' method in orgparse, optional dependency


@cdf
def cross_trainer_manual_dataframe() -> DataFrameT:
    '''
    Only manual org-mode entries
    '''
    import pandas as pd
    df = pd.DataFrame(cross_trainer_data())
    return df

# this should be enough?..
_DELTA = timedelta(hours=10)

# todo check error handling by introducing typos (e.g. especially dates) in org-mode
@cdf
def dataframe() -> DataFrameT:
    '''
    Attaches manually logged data (which Endomondo can't capture) and attaches it to Endomondo
    '''
    import pandas as pd

    from ...endomondo import dataframe as EDF
    edf = EDF()
    edf = edf[edf['sport'].str.contains('Cross training')]

    mdf = cross_trainer_manual_dataframe()
    # TODO shit. need to always remember to split errors???
    # on the other hand, dfs are always untyped. so it's not too bad??
    # now for each manual entry, find a 'close enough' endomondo entry
    # ideally it's a 1-1 (or 0-1) relationship, but there might be errors
    rows = []
    idxs = [] # type: ignore[var-annotated]
    NO_ENDOMONDO = 'no endomondo matches'
    for _i, row in mdf.iterrows():
        rd = row.to_dict()
        mdate = row['date']
        if pd.isna(mdate):
            # todo error handling got to be easier. seriously, mypy friendly dataframes would be amazing
            idxs.append(None)
            rows.append(rd) # presumably has an error set
            continue

        idx: int | None
        close = edf[edf['start_time'].apply(lambda t: pd_date_diff(t, mdate)).abs() < _DELTA]
        if len(close) == 0:
            idx = None
            d = {
                **rd,
                'error': NO_ENDOMONDO,
            }
        elif len(close) > 1:
            idx = None
            d = {
                **rd,
                'error': f'one manual, many endomondo: {close}',
            }
        else:
            idx = close.index[0]
            d = rd

            if idx in idxs:
                # todo might be a good idea to remove the original match as well?
                idx = None
                d = {
                    **rd,
                    'error': 'one endomondo, many manual',
                }
        idxs.append(idx)
        rows.append(d)
    mdf = pd.DataFrame(rows, index=idxs)

    # todo careful about 'how'? we need it to preserve the errors
    # maybe pd.merge is better suited for this??
    df = edf.join(mdf, how='outer', rsuffix='_manual')
    # todo reindex? so we don't have Nan leftovers

    # todo set date anyway? maybe just squeeze into the index??
    noendo = df['error'] == NO_ENDOMONDO
    # meh. otherwise the column type ends up object
    tz = df[noendo]['start_time'].dtype.tz
    df.loc[noendo, 'start_time'    ] = df[noendo]['date'           ].dt.tz_convert(tz)
    df.loc[noendo, 'duration'      ] = df[noendo]['duration_manual']
    df.loc[noendo, 'heart_rate_avg'] = df[noendo]['hr_avg'         ]

    # todo set sport?? set source?
    return df
# TODO arbitrate kcal, duration, avg hr
# compare power and hr? add 'quality' function??
# TODO wtf?? where is speed coming from??


from ...core import Stats, stat


def stats() -> Stats:
    return stat(cross_trainer_data)


def compare_manual() -> None:
    df = dataframe()
    df = df.set_index('start_time')

    df = df[[
        'kcal'    , 'kcal_manual',
        'duration', 'duration_manual',
    ]].dropna()
    print(df.to_string())


def pd_date_diff(a, b) -> timedelta:
    # ugh. pandas complains when we subtract timestamps in different timezones
    assert a.tzinfo is not None, a
    assert b.tzinfo is not None, b
    return a.to_pydatetime() - b.to_pydatetime()



from __future__ import annotations

import json

from . import warnings
from .pytest import parametrize


def json_loads(data: bytes | str):
    """
    Tries loading with orjson if it's available, otherwise falls back to stdlib json.
    """
    try:
        import orjson
    except ModuleNotFoundError as e:
        if e.name != 'orjson':
            raise
        warnings.medium("orjson is not installed, falling back to stdlib json. Install orjson for better performance.")
        return json.loads(data)
    else:
        return orjson.loads(data)


@parametrize("data, description", [('{"key": "value"}', "str"), (b'{"key": "value"}', "bytes")])
def test_json_loads_stdlib(data: str | bytes, description: str) -> None:
    from unittest.mock import patch

    import pytest

    with patch.dict('sys.modules', {'orjson': None}):  # hide orjson if it's installed
        with pytest.warns(UserWarning, match="orjson is not installed"):
            result = json_loads(data)

    assert result == {"key": "value"}


@parametrize("data, description", [('{"key": "value"}', "str"), (b'{"key": "value"}', "bytes")])
def test_json_loads_orjson(data: str | bytes, description: str) -> None:
    import pytest

    try:
        import orjson
    except ModuleNotFoundError:
        pytest.fail("orjson is not installed")

    result = json_loads(data)

    assert result == {"key": "value"}



from __future__ import annotations

import sys
import types
from typing import Any


# The idea behind this one is to support accessing "overlaid/shadowed" modules from namespace packages
# See usage examples here:
# - https://github.com/karlicoss/hpi-personal-overlay/blob/master/src/my/util/hpi_heartbeat.py
# - https://github.com/karlicoss/hpi-personal-overlay/blob/master/src/my/twitter/all.py
# Suppose you want to use my.twitter.talon, which isn't in the default all.py
# You could just copy all.py to your personal overlay, but that would mean duplicating
# all the code and possible upstream changes.
# Alternatively, you could import the "original" my.twitter.all module from "overlay" my.twitter.all
# _ORIG = import_original_module(__name__, __file__)
# this would magically take care of package import path etc,
# and should import the "original" my.twitter.all as _ORIG
# After that you can call its methods, extend etc.
def import_original_module(
    c__module__: str,
    c__file__: str,
    *,
    star: bool = False,
    globals: dict[str, Any] | None = None,  # noqa: A002
) -> types.ModuleType:
    """
    :param c__module__: __module__ of the callee.
    :param c__file__: __file__ of the callee
    :param start: if True, do 'import *' from the original module (into 'globals' dict, which needs to be set)
    :param globals: if star is True, this is the globals dict to update
    """
    if star:
        assert globals is not None, "globals must be set if star is True"

    module_to_restore = sys.modules[c__module__]

    # NOTE: we really wanna to hack the actual package of the module
    # rather than just top level my.
    # since that would be a bit less disruptive
    module_pkg = module_to_restore.__package__
    assert module_pkg is not None
    parent = sys.modules[module_pkg]

    my_path = parent.__path__._path  # type: ignore[attr-defined]
    my_path_orig = list(my_path)

    def fixup_path() -> None:
        for p in my_path_orig:
            starts = c__file__.startswith(p)
            if starts:
                my_path.remove(p)
        # should remove at least one item
        # sometimes it can be >1 when pytest is involved or some other path madness
        assert len(my_path) < len(my_path_orig), (my_path, my_path_orig, c__file__)

    try:
        # first, prepare module path to perform the import
        fixup_path()
        try:
            del sys.modules[c__module__]
            # NOTE: we're using __import__ instead of importlib.import_module
            # since it's closer to the actual normal import (e.g. imports subpackages etc properly )
            # fromlist=[None] forces it to return rightmost child
            # (otherwise would just return 'my' package)
            res = __import__(c__module__, fromlist=[None])  # type: ignore[list-item]
            if star:
                assert globals is not None
                globals.update({k: v for k, v in vars(res).items() if not k.startswith('_')})
            return res
        finally:
            sys.modules[c__module__] = module_to_restore
    finally:
        my_path[:] = my_path_orig



"""
Zulip data from [[https://memex.zulipchat.com/help/export-your-organization][Organization export]]
"""

from __future__ import annotations

import json
from abc import abstractmethod
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from datetime import datetime, timezone
from itertools import count
from pathlib import Path

from my.core import (
    Json,
    Paths,
    Res,
    Stats,
    assert_never,
    datetime_aware,
    get_files,
    make_logger,
    stat,
    warnings,
)

logger = make_logger(__name__)


class config:
    @property
    @abstractmethod
    def export_path(self) -> Paths:
        """paths[s]/glob to the exported JSON data"""
        raise NotImplementedError


def make_config() -> config:
    from my.config import zulip as user_config

    class combined_config(user_config.organization, config):
        pass

    return combined_config()


def inputs() -> Sequence[Path]:
    # TODO: seems like export ids are kinda random..
    # not sure what's the best way to figure out the last without renaming?
    # could use mtime perhaps?
    return get_files(make_config().export_path, sort=False)


@dataclass(frozen=True)
class Server:
    id: int
    string_id: str
    name: str


@dataclass(frozen=True)
class Sender:
    id: int
    # todo make optional?
    full_name: str
    email: str


# from the data, seems that subjects are completely implicit and determined by name?
# streams have ids (can extract from realm/zerver_stream), but unclear how to correlate messages/topics to streams?
@dataclass(frozen=True)
class _Message:
    # todo hmm not sure what would be a good field order..
    id: int
    sent: datetime_aware  # double checked and they are in utc
    subject: str
    sender_id: int
    server_id: int
    content: str  # TODO hmm, it keeps markdown, not sure how/whether it's worth to prettify at all?
    # TODO recipient??
    # todo keep raw item instead? not sure


@dataclass(frozen=True)
class Message:
    id: int
    sent: datetime_aware
    subject: str
    sender: Sender
    server: Server
    content: str

    @property
    def permalink(self) -> str:
        # seems that these link to the same message
        # https://memex.zulipchat.com/#narrow/stream/284580-python/topic/py-spy.20profiler/near/234798881
        # https://memex.zulipchat.com/#narrow/stream/284580/near/234798881
        # https://memex.zulipchat.com/#narrow/near/234798881
        # however not sure how to correlate stream id and message/topic for now, so preferring the latter version
        return f'https://{self.server.string_id}.zulipchat.com/#narrow/near/{self.id}'


# todo cache it
def _entities() -> Iterator[Res[Server | Sender | _Message]]:
    last = max(inputs())

    logger.info(f'extracting data from {last}')

    root: Path | None = None

    if last.is_dir():  # if it's already CPath, this will match it
        root = last
    else:
        try:
            from kompress import CPath

            root = CPath(last)
            assert len(list(root.iterdir())) > 0  # trigger to check if we have the kompress version with targz support
        except Exception as e:
            logger.exception(e)
            warnings.high("Upgrade 'kompress' to latest version with native .tar.gz support. Falling back to unpacking to tmp dir.")

    if root is None:
        from my.core.structure import match_structure

        with match_structure(last, expected=()) as res:  # expected=() matches it regardless any patterns
            [root] = res
            yield from _process_one(root)
    else:
        yield from _process_one(root)


def _process_one(root: Path) -> Iterator[Res[Server | Sender | _Message]]:
    [subdir] = root.iterdir()  # there is a directory inside tar file, first name should be that

    rj = json.loads((subdir / 'realm.json').read_text())

    [sj] = rj['zerver_realm']
    server = Server(
        id=sj['id'],
        string_id=sj['string_id'],
        name=sj['name'],
    )
    yield server

    for j in rj['zerver_userprofile']:
        yield Sender(
            id=j['id'],
            full_name=j['full_name'],
            email=j['email'],
        )

    for j in rj['zerver_userprofile_crossrealm']:  # e.g. zulip bot
        yield Sender(
            id=j['id'],
            full_name=j['email'],  # doesn't seem to have anything
            email=j['email'],
        )

    def _parse_message(j: Json) -> _Message:
        ds = j['date_sent']
        # fmt: off
        return _Message(
            id        = j['id'],
            sent      = datetime.fromtimestamp(ds, tz=timezone.utc),
            subject   = j['subject'],
            sender_id = j['sender'],
            server_id = server.id,
            content   = j['content'],
        )
        # fmt: on

    for idx in count(start=1, step=1):
        fname = f'messages-{idx:06}.json'
        fpath = subdir / fname
        if not fpath.exists():
            break
        mj = json.loads(fpath.read_text())
        # TODO handle  zerver_usermessage
        for j in mj['zerver_message']:
            try:
                yield _parse_message(j)
            except Exception as e:
                yield e


def messages() -> Iterator[Res[Message]]:
    id2sender: dict[int, Sender] = {}
    id2server: dict[int, Server] = {}
    for x in _entities():
        if isinstance(x, Exception):
            yield x
            continue
        if isinstance(x, Server):
            id2server[x.id] = x
            continue
        if isinstance(x, Sender):
            id2sender[x.id] = x
            continue
        if isinstance(x, _Message):
            # TODO a bit copypasty... wonder if possible to mixin or something instead
            yield Message(
                id=x.id,
                sent=x.sent,
                subject=x.subject,
                sender=id2sender[x.sender_id],
                server=id2server[x.server_id],
                content=x.content,
            )
            continue
        assert_never(x)


def stats() -> Stats:
    return {**stat(messages)}



'''
Module for locating and accessing [[https://takeout.google.com][Google Takeout]] data
'''

from __future__ import annotations

from my.core import __NOT_HPI_MODULE__  # isort: skip

from abc import abstractmethod
from collections.abc import Iterable
from pathlib import Path

from more_itertools import last

from my.core import Paths, get_files


class config:
    """
    path/paths/glob for the takeout zips
    """

    @property
    @abstractmethod
    def takeout_path(self) -> Paths:
        raise NotImplementedError


# TODO rename 'google' to 'takeout'? not sure


def make_config() -> config:
    from my.config import google as user_config

    class combined_config(user_config, config): ...

    return combined_config()


def get_takeouts(*, path: str | None = None) -> Iterable[Path]:
    """
    Sometimes google splits takeout into multiple archives, so we need to detect the ones that contain the path we need
    """
    # TODO zip is not great..
    # allow a lambda expression? that way the user could restrict it
    cfg = make_config()
    for takeout in get_files(cfg.takeout_path, glob='*.zip'):
        if path is None or (takeout / path).exists():
            yield takeout


def get_last_takeout(*, path: str | None = None) -> Path | None:
    return last(get_takeouts(path=path), default=None)


# TODO might be a good idea to merge across multiple takeouts...
# perhaps even a special takeout module that deals with all of this automatically?
# e.g. accumulate, filter and maybe report useless takeouts?



from __future__ import annotations

import shutil
import sqlite3
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Callable, Literal, Union, overload

from .common import PathIsh
from .compat import assert_never


def sqlite_connect_immutable(db: PathIsh) -> sqlite3.Connection:
    return sqlite3.connect(f'file:{db}?immutable=1', uri=True)


def test_sqlite_connect_immutable(tmp_path: Path) -> None:
    db = str(tmp_path / 'db.sqlite')
    with sqlite3.connect(db) as conn:
        conn.execute('CREATE TABLE testtable (col)')

    import pytest

    with pytest.raises(sqlite3.OperationalError, match='readonly database'):
        with sqlite_connect_immutable(db) as conn:
            conn.execute('DROP TABLE testtable')

    # succeeds without immutable
    with sqlite3.connect(db) as conn:
        conn.execute('DROP TABLE testtable')


SqliteRowFactory = Callable[[sqlite3.Cursor, sqlite3.Row], Any]


def dict_factory(cursor, row):
    fields = [column[0] for column in cursor.description]
    return dict(zip(fields, row))


Factory = Union[SqliteRowFactory, Literal['row', 'dict']]


@contextmanager
def sqlite_connection(db: PathIsh, *, immutable: bool = False, row_factory: Factory | None = None) -> Iterator[sqlite3.Connection]:
    dbp = f'file:{db}'
    # https://www.sqlite.org/draft/uri.html#uriimmutable
    if immutable:
        # assert results in nicer error than sqlite3.OperationalError
        assert Path(db).exists(), db
        dbp = f'{dbp}?immutable=1'
    row_factory_: Any = None
    if row_factory is not None:
        if callable(row_factory):
            row_factory_ = row_factory
        elif row_factory == 'row':
            row_factory_ = sqlite3.Row
        elif row_factory == 'dict':
            row_factory_ = dict_factory
        else:
            assert_never()

    conn = sqlite3.connect(dbp, uri=True)
    try:
        conn.row_factory = row_factory_
        with conn:
            yield conn
    finally:
        # Connection context manager isn't actually closing the connection, only keeps transaction
        conn.close()


# TODO come up with a better name?
# NOTE: this is tested by tests/sqlite.py::test_sqlite_read_with_wal
def sqlite_copy_and_open(db: PathIsh) -> sqlite3.Connection:
    """
    'Snapshots' database and opens by making a deep copy of it including journal/WAL files
    """
    dp = Path(db)
    # TODO make atomic/check mtimes or something
    dest = sqlite3.connect(':memory:')
    with TemporaryDirectory() as td:
        tdir = Path(td)
        # shm should be recreated from scratch -- safer not to copy perhaps
        tocopy = [dp] + [p for p in dp.parent.glob(dp.name + '-*') if not p.name.endswith('-shm')]
        for p in tocopy:
            shutil.copy(p, tdir / p.name)
        with sqlite3.connect(str(tdir / dp.name)) as conn:
            conn.backup(target=dest)
        conn.close()
    return dest


# NOTE hmm, so this kinda works
# V = TypeVar('V', bound=Tuple[Any, ...])
# def select(cols: V, rest: str, *, db: sqlite3.Connection) -> Iterator[V]:
# but sadly when we pass columns (Tuple[str, ...]), it seems to bind this type to V?
# and then the return type ends up as Iterator[Tuple[str, ...]], which isn't desirable :(
# a bit annoying to have this copy-pasting, but hopefully not a big issue

# fmt: off
@overload
def select(cols: tuple[str                                   ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[tuple[Any                                   ]]: ...
@overload
def select(cols: tuple[str, str                              ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[tuple[Any, Any                              ]]: ...
@overload
def select(cols: tuple[str, str, str                         ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[tuple[Any, Any, Any                         ]]: ...
@overload
def select(cols: tuple[str, str, str, str                    ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[tuple[Any, Any, Any, Any                    ]]: ...
@overload
def select(cols: tuple[str, str, str, str, str               ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[tuple[Any, Any, Any, Any, Any               ]]: ...
@overload
def select(cols: tuple[str, str, str, str, str, str          ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[tuple[Any, Any, Any, Any, Any, Any          ]]: ...
@overload
def select(cols: tuple[str, str, str, str, str, str, str     ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[tuple[Any, Any, Any, Any, Any, Any, Any     ]]: ...
@overload
def select(cols: tuple[str, str, str, str, str, str, str, str], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[tuple[Any, Any, Any, Any, Any, Any, Any, Any]]: ...
# fmt: on

def select(cols, rest, *, db):
    # db arg is last cause that results in nicer code formatting..
    return db.execute('SELECT ' + ','.join(cols) + ' ' + rest)


class SqliteTool:
    def __init__(self, connection: sqlite3.Connection) -> None:
        self.connection = connection

    def get_db_path(self) -> str:
        return self.connection.execute('PRAGMA database_list').fetchall()[0][2]

    def _get_sqlite_master(self) -> dict[str, str]:
        res = {}
        for c in self.connection.execute('SELECT name, type FROM sqlite_master'):
            [name, type_] = c
            assert type_ in {'table', 'index', 'view', 'trigger'}, (name, type_)  # just in case
            res[name] = type_
        return res

    def get_table_names(self) -> list[str]:
        master = self._get_sqlite_master()
        res = []
        for name, type_ in master.items():
            if type_ != 'table':
                continue
            res.append(name)
        return res

    def get_table_schema(self, name: str) -> dict[str, str]:
        """
        Returns map from column name to column type

        NOTE: Sometimes this doesn't work if the db has some extensions (e.g. happens for facebook apps)
              In this case you might still be able to use get_table_names
        """
        schema: dict[str, str] = {}
        for row in self.connection.execute(f'PRAGMA table_info(`{name}`)'):
            col   = row[1]
            type_ = row[2]
            # hmm, somewhere between 3.34.1 and 3.37.2, sqlite started normalising type names to uppercase
            # let's do this just in case since python < 3.10 are using the old version
            # e.g. it could have returned 'blob' and that would confuse blob check (see _check_allowed_blobs)
            type_ = type_.upper()
            schema[col] = type_
        return schema

    def get_table_schemas(self) -> dict[str, dict[str, str]]:
        return {name: self.get_table_schema(name) for name in self.get_table_names()}



"""
Extracts data from the official Google Maps app for Android (uses gmm_sync.db for now)
"""
from __future__ import annotations

REQUIRES = [
    "protobuf",  # for parsing blobs from the database
]

from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any
from urllib.parse import quote

from my.core import LazyLogger, Paths, Res, datetime_aware, get_files
from my.core.common import unique_everseen
from my.core.sqlite import sqlite_connection

from ._android_protobuf import parse_labeled, parse_list, parse_place

import my.config  # isort: skip

logger = LazyLogger(__name__)


@dataclass
class config(my.config.google.maps.android):
    # paths[s]/glob to the exported sqlite databases
    export_path: Paths


def inputs() -> Sequence[Path]:
    # TODO note sure if need to use all dbs? possibly the last one contains everything?
    return get_files(config.export_path)


PlaceId = str
ListId = str
ListName = str


@dataclass(eq=True, frozen=True)
class Location:
    lat: float
    lon: float

    @property
    def url(self) -> str:
        return f'https://maps.google.com/?q={self.lat},{self.lon}'


@dataclass(unsafe_hash=True)
class Place:
    id: PlaceId
    list_name: ListName  # TODO maybe best to keep list id?
    created_at: datetime_aware  # TODO double check it's utc?
    updated_at: datetime_aware  # TODO double check it's utc?
    title: str
    location: Location
    address: str | None
    note: str | None

    @property
    def place_url(self) -> str:
        title = quote(self.title)
        return f'https://www.google.com/maps/place/{title}/data=!4m2!3m1!1s{self.id}'

    @property
    def location_url(self) -> str:
        return self.location.url


def _process_one(f: Path):
    with sqlite_connection(f, row_factory='row') as conn:
        msg: Any

        lists: dict[ListId, ListName] = {}
        for row in conn.execute('SELECT * FROM sync_item_data WHERE corpus == 13'):  # 13 looks like lists (e.g. saved/favorited etc)
            server_id = row['server_id']

            if server_id is None:
                # this is the case for Travel plans, Followed places, Offers
                # todo alternatively could use string_index column instead maybe?
                continue

            blob = row['item_proto']
            msg = parse_list(blob)
            name = msg.f1.name
            lists[server_id] = name

        for row in conn.execute('SELECT * FROM sync_item_data WHERE corpus == 11'):  # this looks like 'Labeled' list
            ts = row['timestamp'] / 1000
            created = datetime.fromtimestamp(ts, tz=timezone.utc)

            server_id = row['server_id']
            [item_type, item_id] = server_id.split(':')
            if item_type != '3':
                # the ones that are not 3 are home/work address?
                continue

            blob = row['item_proto']
            msg = parse_labeled(blob)
            address = msg.address.full
            if address == '':
                address = None

            location = Location(lat=row['latitude_e6'] / 1e6, lon=row['longitude_e6'] / 1e6)

            yield Place(
                id=item_id,
                list_name='Labeled',
                created_at=created,
                updated_at=created,  # doesn't look like it has 'updated'?
                title=msg.title,
                location=location,
                address=address,
                note=None,  # don't think these allow notes
            )

        for row in conn.execute('SELECT * FROM sync_item_data WHERE corpus == 14'):  # this looks like actual individual places
            server_id = row['server_id']
            [list_id, _, id1, id2] = server_id.split(':')
            item_id = f'{id1}:{id2}'

            list_name = lists[list_id]

            blob = row['item_proto']
            msg = parse_place(blob)
            title = msg.f1.title
            note = msg.f1.note
            if note == '':  # seems that protobuf does that?
                note = None

            # TODO double check timezone
            created = datetime.fromtimestamp(msg.f1.created.seconds, tz=timezone.utc).replace(microsecond=msg.f1.created.nanos // 1000)

            # NOTE: this one seems to be the same as row['timestamp']
            updated = datetime.fromtimestamp(msg.f1.updated.seconds, tz=timezone.utc).replace(microsecond=msg.f1.updated.nanos // 1000)

            address = msg.f2.addr1  # NOTE: there is also addr2, but they seem identical :shrug:
            if address == '':
                address = None

            location = Location(lat=row['latitude_e6'] / 1e6, lon=row['longitude_e6'] / 1e6)

            place = Place(
                id=item_id,
                list_name=list_name,
                created_at=created,
                updated_at=updated,
                title=title,
                location=location,
                address=address,
                note=note,
            )

            # ugh. in my case it's violated by one place by about 1 second??
            # assert place.created_at <= place.updated_at
            yield place


def saved() -> Iterator[Res[Place]]:
    def it() -> Iterator[Res[Place]]:
        paths = inputs()
        total = len(paths)
        width = len(str(total))
        for idx, path in enumerate(paths):
            logger.info(f'processing [{idx:>{width}}/{total:>{width}}] {path}')
            yield from _process_one(path)
    return unique_everseen(it)


# Summary of databases on Android (as of 20240101)
# -1_optimized_threads.notifications.db -- empty
# 1_optimized_threads.notifications.db  -- empty
# 1_tasks.notifications.db              -- empty
# -1_threads.notifications.db           -- empty
# 1_threads.notifications.db            -- doesn't look like anything interested, some trip anniversaries etc?
# 1_thread_surveys.notifications.db     -- empty
# 2_threads.notifications.db            -- empty
# accounts.notifications.db             -- just one row with account id
# brella_example_store                  -- empty
# gmm_myplaces.db                       -- contains just a few places? I think it's a subset of "Labeled"
# gmm_storage.db                        -- pretty huge, like 50Mb. I suspect it contains cache for places on maps or something
# gmm_sync.db                           -- processed above
# gnp_fcm_database                      -- list of accounts
# google_app_measurement_local.db       -- empty
# inbox_notifications.db                -- nothing interesting
# <email>_room_notifications.db         -- trip anniversaties?
# lighter_messaging_1.db                -- empty
# lighter_messaging_2.db                -- empty
# lighter_registration.db               -- empty
# peopleCache_<email>_com.google_14.db  -- contacts cache or something
# portable_geller_<email>.db            -- looks like analytics
# primes_example_store                  -- looks like analytics
# pseudonymous_room_notifications.db    -- looks like analytics
# ue3.db                                -- empty
# ugc_photos_location_data.db           -- empty
# ugc-sync.db                           -- empty
# updates-tab-visit.db                  -- empty



# todo move to hpi core? depends if it's useful for other modules
from __future__ import annotations

from typing import Any, Callable


def dict_diff(a: dict, b: dict) -> dict:
    """
    >>> dict_diff({'d': 1, 'a': 5, 'b': 2}, {'d': 2, 'a': 5, 'c': 4})
    {'d': (1, 2), 'b': (2, None), 'c': (None, 4)}
    """
    diff = {}
    for key, av in a.items():
        if key not in b:
            diff[key] = (av, None)
        else:
            bv = b[key]
            if av != bv:
                diff[key] = (av, bv)

    for key, bv in b.items():
        if key not in a:
            diff[key] = (None, bv)
        # otherwise already handled in other loop
    return diff


# todo maybe move to .core.sqlite? not sure
DbRow = dict[str, Any]


_I = dict[str, Any]


class MultiKeyTracker:
    """
    Helper to find equivalence between dicts that don't have a unique key, but multiple attributes that might overlap
    """

    def __init__(
        self,
        *,
        keys: list[tuple[str, Any]],
        updater: Callable[[_I, _I], None],
    ) -> None:
        self.items: list[tuple[Any, _I]] = []
        self.keys = keys
        self.updater = updater

    def _key(self, item: _I) -> set:
        r = {}
        for k, invalid in self.keys:
            kv = item.get(k)
            if kv in {invalid, None}:
                continue
            r[k] = kv
        return set(r.items())

    def set(self, item: _I, *, add: bool, update: bool) -> _I | None:  # FIXME add types
        rk = self._key(item)
        for kk, v in self.items:
            if len(kk & rk) == 0:
                continue
            # found equivalent item
            if update:
                self.updater(v, item)
                kk.update(rk)
            return v  # return previously present row
        if add:
            self.items.append((rk, item))
        return None

    def get(self, item: _I) -> _I:
        res = self.set(item, add=False, update=False)
        assert res is not None
        return res



"""
[[https://roamresearch.com][Roam]] data
"""
from __future__ import annotations

import re
from collections.abc import Iterator
from datetime import datetime, timezone
from itertools import chain
from pathlib import Path
from typing import NamedTuple

from my.config import roamresearch as config

from .core import Json, LazyLogger, get_files

logger = LazyLogger(__name__)


def last() -> Path:
    return max(get_files(config.export_path))


class Keys:
    CREATED    = 'create-time'
    EDITED     = 'edit-time'
    EDIT_EMAIL = 'edit-email'
    STRING     = 'string'
    CHILDREN   = 'children'
    TITLE      = 'title'
    UID        = 'uid'


class Node(NamedTuple):
    raw: Json

    # TODO not sure if UTC
    @property
    def created(self) -> datetime:
        ct = self.raw.get(Keys.CREATED)
        if ct is not None:
            return datetime.fromtimestamp(ct / 1000, tz=timezone.utc)
        # ugh. daily notes don't have create time for some reason???

        title = self.title
        if title is None:
            return self.edited # fallback TODO log?
        # the format is 'February 8th, 2020'. Fucking hell.
        m = re.fullmatch(r'(\w+) (\d+)\w+, (\d+)', title)
        if m is None:
            return self.edited # fallback TODO log?
        # strip off 'th'/'rd' crap
        dts = m.group(1) + ' ' + m.group(2) + ' ' + m.group(3)
        dt = datetime.strptime(dts, '%B %d %Y').replace(tzinfo=timezone.utc)
        return dt

    @property
    def edited(self) -> datetime:
        rt = self.raw[Keys.EDITED]
        return datetime.fromtimestamp(rt / 1000, tz=timezone.utc)

    @property
    def title(self) -> str | None:
        return self.raw.get(Keys.TITLE)

    @property
    def body(self) -> str | None:
        return self.raw.get(Keys.STRING)

    @property
    def children(self) -> list[Node]:
        # TODO cache? needs a key argument (because of Json)
        ch = self.raw.get(Keys.CHILDREN, [])
        return list(map(Node, ch))

    @property
    def path(self) -> str:
        username = config.username # sadly, Roam research export doesn't provide it
        return f'{username}/page/{self.uid}'

    @property
    def permalink(self) -> str:
        return f'https://roamresearch.com/#/app/{self.path}'

    @property
    def uid(self) -> str:
        u = self.raw.get(Keys.UID)
        if u is not None:
            return u
        # ugh. so None apparently means "Daily note"

        # yes, it is using US date format...
        return self.created.strftime('%m-%d-%Y')

    def empty(self) -> bool:
        # sometimes nodes are empty. two cases:
        # - no heading -- child notes, like accidental enter presses I guess
        # - heading    -- notes that haven't been created yet
        return len(self.body or '') == 0 and len(self.children) == 0

    def traverse(self) -> Iterator[Node]:
        # not sure about __iter__, because might be a bit unintuitive that it's recursive..
        yield self
        for c in self.children:
            yield from c.traverse()

    def _render(self) -> Iterator[str]:
        ss = f'[{self.created:%Y-%m-%d %H:%M}] {self.title or " "}'
        body = self.body
        sc = chain.from_iterable(c._render() for c in self.children)

        yield ss
        if body is not None:
            yield body
        yield self.permalink
        for c in sc:
            yield '|  ' + c

    def render(self) -> str:
        return '\n'.join(self._render())

    def __repr__(self):
        return f'Node(created={self.created}, title={self.title}, body={self.body})'

    @staticmethod
    def make(raw: Json) -> Iterator[Node]:
        is_empty = set(raw.keys()) == {Keys.EDITED, Keys.EDIT_EMAIL, Keys.TITLE}
        # not sure about that... but daily notes end up like that
        if is_empty:
            # todo log?
            return
        yield Node(raw)


class Roam:
    def __init__(self, raw: list[Json]) -> None:
        self.raw = raw

    @property
    def notes(self) -> list[Node]:
        return list(chain.from_iterable(map(Node.make, self.raw)))

    def traverse(self) -> Iterator[Node]:
        for n in self.notes:
            yield from n.traverse()


def roam() -> Roam:
    import json
    raw = json.loads(last().read_text())
    roam = Roam(raw)
    return roam


def print_all_notes():
    # just a demo method
    # TODO demonstrate dumping as org-mode??
    for n in roam().notes:
        print(n.render())

# TODO could generate org-mode mirror in a single file for a demo?



import os
from pathlib import Path

import pytest

V = 'HPI_TESTS_KARLICOSS'

skip_if_not_karlicoss = pytest.mark.skipif(
    V not in os.environ,
    reason=f'test only works on @karlicoss data for now. Set env variable {V}=true to override.',
)


def hpi_repo_root() -> Path:
    root_dir = Path(__file__).absolute().parent.parent.parent.parent
    src_dir = root_dir / 'src'
    assert src_dir.exists(), src_dir
    return root_dir


def testdata() -> Path:
    d = hpi_repo_root() / 'testdata'
    assert d.exists(), d
    return d


# prevent pytest from treating this as test
testdata.__test__ = False  # type: ignore[attr-defined]



'''
Simple location provider, serving as a fallback when more detailed data isn't available
'''

from __future__ import annotations

from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from datetime import datetime, time, timezone
from functools import cache
from typing import cast

from my.config import location as user_config
from my.location.common import DateIsh, LatLon
from my.location.fallback.common import DateExact, FallbackLocation


@dataclass
class Config(user_config):
    home: LatLon | Sequence[tuple[DateIsh, LatLon]]

    # default ~30km accuracy
    # this is called 'home_accuracy' since it lives on the base location.config object,
    # to differentiate it from accuracy for other providers
    home_accuracy: float = 30_000.0

    # TODO could make current Optional and somehow determine from system settings?
    @property
    def _history(self) -> Sequence[tuple[datetime, LatLon]]:
        home1 = self.home
        # todo ugh, can't test for isnstance LatLon, it's a tuple itself
        home2: Sequence[tuple[DateIsh, LatLon]]
        if isinstance(home1[0], tuple):
            # already a sequence
            home2 = cast(Sequence[tuple[DateIsh, LatLon]], home1)
        else:
            # must be a pair of coordinates. also doesn't really matter which date to pick?
            loc = cast(LatLon, home1)
            home2 = [(datetime.min, loc)]

        # todo cache?
        res = []
        for x, loc in home2:
            dt: datetime
            if isinstance(x, str):
                dt = datetime.fromisoformat(x)
            elif isinstance(x, datetime):
                dt = x
            else:
                dt = datetime.combine(x, time.min)
            # todo not sure about doing it here, but makes it easier to compare..
            if dt.tzinfo is None:
                dt = dt.replace(tzinfo=timezone.utc)
            res.append((dt, loc))
        res = sorted(res, key=lambda p: p[0])
        return res


from ...core.cfg import make_config

config = make_config(Config)


@cache
def get_location(dt: datetime) -> LatLon:
    '''
    Interpolates the location at dt
    '''
    loc = list(estimate_location(dt))
    assert len(loc) == 1
    return loc[0].lat, loc[0].lon


# TODO: in python3.8, use functools.cached_property instead?
@cache
def homes_cached() -> list[tuple[datetime, LatLon]]:
    return list(config._history)


def estimate_location(dt: DateExact) -> Iterator[FallbackLocation]:
    from my.location.fallback.common import _datetime_timestamp
    d: float = _datetime_timestamp(dt)
    hist = list(reversed(homes_cached()))
    for pdt, (lat, lon) in hist:
        if d >= pdt.timestamp():
            yield FallbackLocation(
                lat=lat,
                lon=lon,
                accuracy=config.home_accuracy,
                dt=datetime.fromtimestamp(d, timezone.utc),
                datasource='via_home')
            return

    # I guess the most reasonable is to fallback on the first location
    lat, lon = hist[-1][1]
    yield FallbackLocation(
        lat=lat,
        lon=lon,
        accuracy=config.home_accuracy,
        dt=datetime.fromtimestamp(d, timezone.utc),
        datasource='via_home')



"""
[[https://shop-eu.emfit.com/products/emfit-qs][Emfit QS]] sleep tracker

Consumes data exported by https://github.com/karlicoss/emfitexport
"""

from __future__ import annotations

REQUIRES = [
    'git+https://github.com/karlicoss/emfitexport',
]

import dataclasses
import inspect
from collections.abc import Iterable, Iterator
from contextlib import contextmanager
from datetime import datetime, time, timedelta
from pathlib import Path
from typing import Any

import emfitexport.dal as dal

from my.core import (
    Res,
    Stats,
    get_files,
    stat,
)
from my.core.cachew import cache_dir, mcachew
from my.core.error import extract_error_datetime, set_error_datetime
from my.core.pandas import DataFrameT

from my.config import emfit as config  # isort: skip


Emfit = dal.Emfit


# TODO move to common?
def dir_hash(path: Path):
    mtimes = tuple(p.stat().st_mtime for p in get_files(path, glob='*.json'))
    return mtimes


def _cachew_depends_on():
    return dir_hash(config.export_path)


# TODO take __file__ into account somehow?
@mcachew(cache_path=cache_dir() / 'emfit.cache', depends_on=_cachew_depends_on)
def datas() -> Iterable[Res[Emfit]]:
    # data from emfit is coming in UTC. There is no way (I think?) to know the 'real' timezone, and local times matter more for sleep analysis
    # TODO actually this is wrong?? there is some sort of local offset in the export
    emfit_tz = config.timezone

    ## backwards compatibility (old DAL didn't have cpu_pool argument)
    cpu_pool_arg = 'cpu_pool'
    pass_cpu_pool = cpu_pool_arg in inspect.signature(dal.sleeps).parameters
    if pass_cpu_pool:
        from my.core._cpu_pool import get_cpu_pool

        kwargs = {cpu_pool_arg: get_cpu_pool()}
    else:
        kwargs = {}
    ##

    for x in dal.sleeps(config.export_path, **kwargs):
        if isinstance(x, Exception):
            yield x
        else:
            if x.sid in config.excluded_sids:
                # TODO should be responsibility of export_path (similar to HPI?)
                continue
            # TODO maybe have a helper to 'patch up' all dattetimes in a namedtuple/dataclass?
            # TODO do the same for jawbone data?
            # fmt: off
            x = dataclasses.replace(
                x,
                start      =x.start      .astimezone(emfit_tz),
                end        =x.end        .astimezone(emfit_tz),
                sleep_start=x.sleep_start.astimezone(emfit_tz),
                sleep_end  =x.sleep_end  .astimezone(emfit_tz),
            )
            # fmt: on
            yield x


# TODO should be used for jawbone data as well?
def pre_dataframe() -> Iterable[Res[Emfit]]:
    # TODO shit. I need some sort of interrupted sleep detection?
    g: list[Emfit] = []

    def flush() -> Iterable[Res[Emfit]]:
        if len(g) == 0:
            return
        elif len(g) == 1:
            r = g[0]
            g.clear()
            yield r
        else:
            err = RuntimeError(f'Multiple sleeps per night, not supported yet: {g}')
            set_error_datetime(err, dt=datetime.combine(g[0].date, time.min))
            g.clear()
            yield err

    for x in datas():
        if isinstance(x, Exception):
            yield x
            continue
        # otherwise, Emfit
        if len(g) != 0 and g[-1].date != x.date:
            yield from flush()
        g.append(x)
    yield from flush()


def dataframe() -> DataFrameT:
    dicts: list[dict[str, Any]] = []
    last: Emfit | None = None
    for s in pre_dataframe():
        d: dict[str, Any]
        if isinstance(s, Exception):
            edt = extract_error_datetime(s)
            d = {
                'date': edt,
                'error': str(s),
            }
        else:
            dd = s.date
            pday = dd - timedelta(days=1)
            if last is None or last.date != pday:
                hrv_change = None
            else:
                # todo it's change during the day?? dunno if reasonable metric
                hrv_change = s.hrv_evening - last.hrv_morning
            # todo maybe changes need to be handled in a more generic way?

            # todo ugh. get rid of hardcoding, just generate the schema automatically
            # TODO use 'workdays' provider....
            # fmt: off
            d = {
                'date'       : dd,

                'sleep_start': s.sleep_start,
                'sleep_end'  : s.sleep_end,
                'bed_time'   : s.time_in_bed, # eh, this is derived frop sleep start / end. should we compute it on spot??

                # these are emfit specific
                'coverage'   : s.sleep_hr_coverage,
                'avg_hr'     : s.measured_hr_avg,
                'hrv_evening': s.hrv_evening,
                'hrv_morning': s.hrv_morning,
                'recovery'   : s.recovery,
                'hrv_change' : hrv_change,
                'respiratory_rate_avg': s.respiratory_rate_avg,
            }
            # fmt: on
            last = s  # meh
        dicts.append(d)

    import pandas as pd

    return pd.DataFrame(dicts)


def stats() -> Stats:
    return stat(pre_dataframe)


@contextmanager
def fake_data(nights: int = 500) -> Iterator:
    from tempfile import TemporaryDirectory

    import pytz

    from my.core.cfg import tmp_config

    with TemporaryDirectory() as td:
        tdir = Path(td)
        gen = dal.FakeData()
        gen.fill(tdir, count=nights)

        class override:
            class emfit:
                export_path = tdir
                excluded_sids = ()
                timezone = pytz.timezone('Europe/London')  # meh

        with tmp_config(modules=__name__, config=override) as cfg:
            yield cfg


# TODO remove/deprecate it? I think used by timeline
def get_datas() -> list[Emfit]:
    # todo ugh. run lint properly
    return sorted(datas(), key=lambda e: e.start)  # type: ignore[arg-type]


# TODO move away old entries if there is a diff??



REQUIRES = ["git+https://github.com/purarue/ipgeocache"]

from my.core.warnings import high

from .fallback.via_ip import *

high("my.location.via_ip is deprecated, use my.location.fallback.via_ip instead")



"""
[[https://github.com/burtonator/polar-bookshelf][Polar]] articles and highlights
"""

from __future__ import annotations

import json
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple, cast

from my.core import (
    Json,
    PathIsh,
    Res,
    datetime_aware,
    get_files,
    make_config,
    make_logger,
)
from my.core.compat import add_note, fromisoformat
from my.core.error import sort_res_by
from my.core.konsume import Wdict, Zoomable, wrap

import my.config  # isort: skip

logger = make_logger(__name__)


# todo use something similar to tz.via_location for config fallback
if not TYPE_CHECKING:
    user_config = getattr(my.config, 'polar', None)
else:
    # mypy can't handle dynamic base classes... https://github.com/python/mypy/issues/2477
    user_config = object

# by default, Polar doesn't need any config, so perhaps makes sense to make it defensive here
if user_config is None:

    class user_config:  # type: ignore[no-redef]
        pass


@dataclass
class polar(user_config):
    '''
    Polar config is optional, you only need it if you want to specify custom 'polar_dir'
    '''

    polar_dir: PathIsh = Path('~/.polar').expanduser()  # noqa: RUF009
    defensive: bool = True  # pass False if you want it to fail faster on errors (useful for debugging)


config = make_config(polar)

# todo not sure where it keeps stuff on Windows?
# https://github.com/burtonator/polar-bookshelf/issues/296


# Ok I guess handling comment-level errors is a bit too much..
Cid = str


class Comment(NamedTuple):
    cid: Cid
    created: datetime_aware
    text: str


Hid = str


class Highlight(NamedTuple):
    hid: Hid
    created: datetime_aware
    selection: str
    comments: Sequence[Comment]
    tags: Sequence[str]
    page: int  # 1-indexed
    color: str | None = None


Uid = str


class Book(NamedTuple):
    created: datetime_aware
    uid: Uid
    path: Path
    title: str | None
    # TODO hmmm. I think this needs to be defensive as well...
    # think about it later.
    items: Sequence[Highlight]

    tags: Sequence[str]

    @property
    def filename(self) -> str:
        # TODO deprecate
        return str(self.path)


Result = Res[Book]


class Loader:
    def __init__(self, p: Path) -> None:
        self.path = p
        self.uid = self.path.parent.name

    def load_item(self, meta: Zoomable) -> Iterable[Highlight]:
        meta = cast(Wdict, meta)
        # TODO this should be destructive zoom?
        meta['notes'].zoom()  # TODO ??? is it deliberate?

        meta['pagemarks'].consume_all()

        if 'notes' in meta:
            # TODO something nicer?
            _notes = meta['notes'].zoom()
        else:
            _notes = []
        # not sure what is 'notes', seemed always empty for me?

        comments = list(meta['comments'].zoom().values()) if 'comments' in meta else []
        meta['questions'].zoom()
        meta['flashcards'].zoom()
        highlights = meta['textHighlights'].zoom()

        # TODO could be useful to at least add a meta bout area highlights/screens
        meta['areaHighlights'].consume_all()
        meta['screenshots'].zoom()
        meta['thumbnails'].zoom()
        if 'readingProgress' in meta:
            meta['readingProgress'].consume_all()

        # TODO want to ignore the whole subtree..
        pi = meta['pageInfo'].zoom()
        page = pi['num'].zoom().value
        if 'dimensions' in pi:
            pi['dimensions'].consume_all()

        # TODO how to make it nicer?
        cmap: dict[Hid, list[Comment]] = {}
        vals = list(comments)
        for v in vals:
            cid = v['id'].zoom()
            v['guid'].zoom()
            # TODO values should probably be checked by flow analysis??
            crt = v['created'].zoom()
            _updated = v['lastUpdated'].zoom()
            content = v['content'].zoom()
            html = content['HTML'].zoom()
            refv = v['ref'].zoom().value
            [_, hlid] = refv.split(':')
            ccs = cmap.get(hlid, [])
            cmap[hlid] = ccs
            comment = Comment(
                cid=cid.value,
                created=fromisoformat(crt.value),
                text=html.value,  # TODO perhaps coonvert from html to text or org?
            )
            ccs.append(comment)
            v.consume()
        for h in list(highlights.values()):
            hid = h['id'].zoom().value
            if hid in cmap:
                comments = cmap[hid]
                del cmap[hid]
            else:
                comments = []

            h['guid'].consume()
            crt = h['created'].zoom().value
            _updated = h['lastUpdated'].zoom().value
            h['rects'].ignore()

            # TODO make it more generic..
            htags: list[str] = []
            if 'tags' in h:
                ht = h['tags'].zoom()
                for _k, v in list(ht.items()):
                    ctag = v.zoom()
                    ctag['id'].consume()
                    ct = ctag['label'].zoom()
                    htags.append(ct.value)

            h['textSelections'].ignore()
            h['notes'].consume()
            h['questions'].consume()
            h['flashcards'].consume()
            color = h['color'].zoom().value
            h['images'].ignore()
            # TODO eh, quite excessive \ns...
            text = h['text'].zoom()['TEXT'].zoom().value

            yield Highlight(
                hid=hid,
                created=fromisoformat(crt),
                selection=text,
                comments=tuple(comments),
                tags=tuple(htags),
                page=page,
                color=color,
            )
            h.consume()

        # TODO when I add defensive error policy, support it
        # if len(cmap) > 0:
        #     raise RuntimeError(f'Unconsumed comments: {cmap}')
        # TODO sort by date?

    def load_items(self, metas: Json) -> Iterable[Highlight]:
        for _p, meta in metas.items():  # noqa: PERF102
            with wrap(meta, throw=not config.defensive) as meta:
                yield from self.load_item(meta)

    def load(self) -> Iterable[Result]:
        logger.info('processing %s', self.path)
        j = json.loads(self.path.read_text())

        # TODO konsume here as well?
        di = j['docInfo']
        added = di['added']
        filename = di['filename']
        title = di.get('title', None)
        tags_dict = di['tags']
        pm = j['pageMetas']  # todo handle this too?

        # todo defensive?
        tags = tuple(t['label'] for t in tags_dict.values())

        path = Path(config.polar_dir) / 'stash' / filename

        yield Book(
            created=fromisoformat(added),
            uid=self.uid,
            path=path,
            title=title,
            items=list(self.load_items(pm)),
            tags=tags,
        )


def iter_entries() -> Iterable[Result]:
    for d in get_files(config.polar_dir, glob='*/state.json'):
        loader = Loader(d)
        try:
            yield from loader.load()
        except Exception as e:
            add_note(e, f'^ while processing {d}')
            yield e


def get_entries() -> list[Result]:
    # sorting by first annotation is reasonable I guess???
    # todo perhaps worth making it a pattern? X() returns iterable, get_X returns reasonably sorted list?
    return list(sort_res_by(iter_entries(), key=lambda e: e.created))


## deprecated
if not TYPE_CHECKING:
    # "deprecate" by hiding from mypy
    Error = Exception  # for backwards compat with Orger; can remove later



from my.core import __NOT_HPI_MODULE__  # isort: skip

from typing import TYPE_CHECKING

from my.core.warnings import high

high("DEPRECATED! Please use my.youtube.takeout instead.")

if not TYPE_CHECKING:
    from my.youtube.takeout import *



'''
Data from Podcast Addict Android app (https://podcastaddict.com/app)
'''

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Iterable, Iterator, Sequence
from dataclasses import dataclass
from datetime import datetime, timezone
from itertools import chain
from pathlib import Path
from typing import Protocol

from my.core import Paths, get_files, make_logger
from my.core.compat import KW_ONLY
from my.core.sqlite import sqlite_connection

from ._utils import DbRow, MultiKeyTracker, dict_diff

logger = make_logger(__name__)


class Config(Protocol):
    @property
    @abstractmethod
    def export_path(self) -> Paths:
        """
        Path/glob[s] to sqlite app database (/data/data/com.bambuna.podcastaddict/databases/podcastAddict.db)
        """
        raise NotImplementedError


# FIXME rename to 'default config'? not sure, might be a bit confusing..
def make_config() -> Config:
    from my.config import podcastaddict as user_config

    class combined_config(user_config.android, Config): ...

    return combined_config()


# todo how to keep consistent with Status class?
_IS_EPISODE_INTERESTING = '''
/* not sure what is this.. perhaps file accidentally opened via podcastaddict? */
episodes.podcast_id >= 0
AND
(
episodes.favorite != 0
OR
episodes.seen_status != 0
OR
episodes.position_to_resume != 0
OR
/* NOTE: might have playbackDate != -1, but position_to_resume etc == 0 */
episodes.playbackDate != -1
)
'''

_EPISODES_QUERY = f'''
SELECT episodes.*, podcasts.name AS podcast_name
FROM episodes JOIN podcasts ON episodes.podcast_id == podcasts._id
WHERE {_IS_EPISODE_INTERESTING}
ORDER BY guid
'''

# just to 'normalize' data, remove volatile/useless fields
_VOLATILE_AND_USELESS_COLUMNS = [
    'rssfeed_duration_ms',
    'thumbnail_id',
    'chapters_extracted',
    'downloaded_status',
    'downloaded_date',
    'downloaded_status_int',
    'local_file_name',
    'donation_url',
    'explicit',
    'iTunesType',
    'seasonNb',
    'episodeNb',
    'seasonName',
    'thumbsRating',
    'alternate_urls',
    'size',

    '_id',  # can change if you reset phone or something??

    'creator',  # we rely on podcast.name anyway
    'server_id',  # can change to -1? not sure what it means

    # weird.. sometimes it changes slightly?
    # seems like better to rely on duration_ms anyway, it's not changing
    'duration',

    'download_url',  # probs not super useful, can just use podcast page url
]  # fmt: skip


# these don't look volatile but also perhaps not much use?
# for now we don't do anything about them
# todo perhaps might be useful for bleanser?
_USELESS_COLUMNS = [
    'podcast_id',
    'comments',
    'thumbnail_url',
    'comment_rss',
    'type',
    'new_status',
    'deleted_status',
    'comments_last_modified',
    'comments_etag',
    'flattr',
    'is_virtual',
    'is_artwork_extracted',
    'virtualPodcastName',
    'normalizedType',
    'hasBeenFlattred',
    'download_error_msg',
    'media_extracted_artwork_id',
    'automatically_shared',
    'transcript_url',
    'chapters_url',
]


@dataclass(**KW_ONLY)
class Podcast:
    row: DbRow

    @property
    def name(self) -> str:
        return self.row['name']

    @property
    def homepage(self) -> str | None:
        return self.row['homepage']

    @property
    def subscribed(self) -> bool:
        return self.row['subscribed_status'] > 0

    @property
    def description_html(self) -> str:
        return self.row['description']


_STATUS_COLUMNS = [
    'playbackDate',
    'position_to_resume',
    'seen_status',
]


@dataclass(**KW_ONLY)
class Status:
    # note: there is also 'playing_status' column, but it seems to be always 0
    row: DbRow

    @property
    def playback_dt(self) -> datetime | None:
        """
        Usually playbackDate is changing when there are other status changes... however not always!
        todo perhaps could take from timestamp/file name? not sure
        """
        pb = self.row.get('playbackDate')
        if pb is None:
            return None
        if pb == -1:
            # todo might be nice to tell apart from None case?
            return None
        return datetime.fromtimestamp(pb / 1000, tz=timezone.utc)

    @property
    def position_to_resume(self) -> int | None:
        return self.row.get('position_to_resume')

    @property
    def seen(self) -> bool | None:
        s = self.row.get('seen_status')
        if s is None:
            return None
        return s == 1


@dataclass(**KW_ONLY)
class Episode:
    row: DbRow
    statuses: list[Status]
    podcast: Podcast

    @property
    def guid(self) -> str:
        return self.row['guid']

    @property
    def name(self) -> str:
        return self.row['name']

    @property
    def url(self) -> str | None:
        return self.row['url']

    @property
    def publication_dt(self) -> datetime:
        # todo not 100% sure if it's UTC?
        # tricky to find out for sure, the app doesn't show podcast publication time..
        return datetime.fromtimestamp(self.row['publication_date'] / 1000, tz=timezone.utc)

    @property
    def short_description(self) -> str:
        # NOTE: # there is also
        # - description column, but it seems to be always empty
        # - content column, but it's HTML, but it can be NULL and quite long.. so not using for now
        # - short_description seems to be always present and contain something meaningful
        return self.row['short_description']

    @property
    def duration_ms(self) -> int:
        # NOTE: there is also 'duration' column (hh:mm:ss string), but seems like it's volatile?
        # so we filter it out
        return self.row['duration_ms']


class Processor:
    def __init__(self, *, config: Config) -> None:
        self.config = config

    # NOTE: putting inputs as a separate method might make it friendlier for hpi stat etc to discover?
    def inputs(self) -> Sequence[Path]:
        return get_files(self.config.export_path)

    def _db_rows(self) -> Iterator[tuple[Iterable[DbRow], Iterable[DbRow]]]:
        db_files = self.inputs()
        total = len(db_files)
        width = len(str(total))

        # NOTE: atm we're not doing any row deduplication between different dbs
        # it doesn't really make much performance difference with all the weird podcast processing

        for idx, db_file in enumerate(db_files):
            logger.info(f'processing [{idx:>{width}}/{total:>{width}}] {db_file}')
            with sqlite_connection(db_file, immutable=True, row_factory='dict') as db:
                db_podcast_id_to_podcast: dict[int, DbRow] = {}
                for row in db.execute('SELECT * FROM podcasts WHERE _id > 0'):
                    db_podcast_id = row.pop('_id')
                    db_podcast_id_to_podcast[db_podcast_id] = row

                episodes_rows: list[DbRow] = []
                for row in db.execute(_EPISODES_QUERY):
                    # remove volatile columns for unique_everseen to work properly down the line
                    # TODO can we do it via sqlite instead?
                    row = {k: v for k, v in row.items() if k not in set(_VOLATILE_AND_USELESS_COLUMNS)}

                    # remove podcast_id since it only makes sense within a single db, so we don't use it by accident
                    db_podcast_id = row.pop('podcast_id')
                    row['podcast'] = db_podcast_id_to_podcast[db_podcast_id]
                    # remove podcast name as well, to avoid volatility in Episode objects
                    # we have reference to podcast object anyway
                    row.pop('podcast_name', None)

                    episodes_rows.append(row)

                podcasts_rows = db_podcast_id_to_podcast.values()
                yield (podcasts_rows, episodes_rows)

    def data(self) -> Iterator[Podcast | Episode]:
        # need to keep track of all podcasts ever emitted before, so we can correctly set is_subscribed attribute
        def podcast_meta_updater(current: DbRow, new: DbRow) -> None:
            # ugh. sometimes homepage disappears for no reason??
            nh = new['homepage']
            if nh is None:
                # just remove it from updates in this case, so original homepage is preserved
                new.pop('homepage', None)
            current.update(new)

        podcasts_tracker = MultiKeyTracker(
            keys=[
                # if any of the keys match, the podcasts are the same
                ('server_id', -1),
                ('iTunesID', ''),
                ('feed_url', None),
                ('name', ''),
            ],
            updater=podcast_meta_updater,
        )

        # maps from episode guid
        all_episodes: dict[str, Episode] = {}

        for podcast_rows, episodes_rows in self._db_rows():
            ###
            # we want to recover the exact object that we keep so we can apply updates
            # so this map is from object id() function
            podcasts_in_current_db: dict[int, DbRow] = {}

            # first process (and emit if necessary interesting podcast object)
            # - ones that we're subscribed to
            # - oncs that we listened to (via episodes)
            for prow in chain(
                (prow for prow in podcast_rows if prow['subscribed_status'] > 0),
                (erow['podcast'] for erow in episodes_rows),
            ):
                prev = podcasts_tracker.set(prow, add=True, update=True)
                if prev is not None:
                    # only updated existing item
                    podcasts_in_current_db[id(prev)] = prev
                else:
                    # prow is the newly inserted item
                    podcasts_in_current_db[id(prow)] = prow
                    # FIXME need to update actual podcast object.. it shouldn't change
                    # for now it's ok, we're keeping dict reference anyway
                    yield Podcast(row=prow)

            for prow in podcast_rows:
                prev = podcasts_tracker.set(prow, add=False, update=True)
                if prev is not None:
                    podcasts_in_current_db[id(prev)] = prev

            # FIXME need to be careful with cachew...
            # i.e. we might update row after it's been cached??
            for _k, v in podcasts_tracker.items:
                if id(v) not in podcasts_in_current_db:
                    # if it's not in the db anymore, we've likely unsubscribed
                    v['subscribed_status'] = 0
            ###

            for erow in episodes_rows:
                row = erow
                guid = row['guid']

                ep = all_episodes.get(guid)
                if ep is None:
                    status = Status(row={k: row[k] for k in _STATUS_COLUMNS})
                    podcast_dict = row['podcast']
                    podcast_dict = podcasts_tracker.get(podcast_dict)
                    ep = Episode(
                        row=row,
                        statuses=[status],
                        podcast=Podcast(row=podcast_dict),  # todo make sure to refer to the same Podcast object...
                    )
                    all_episodes[guid] = ep
                    yield ep
                    continue

                # otherwise try to update the object/compute statuses
                prev_row = ep.row
                diff = dict_diff(prev_row, row)

                # TODO hmm
                # maybe it's better to 'normalise' data and split out status update fields from podcast rows?
                # after that emitting diffs etc might be simpler?

                # TODO maybe update all of them? idk if there is much point trying to tell it apart...
                for key in [
                    'url',  # often legitimately updates, best to use the latest one
                    'content',  # changes often
                    'duration_ms',  # only one instance of it changing (by 1ms??) but whatever
                    'categories',  # had one instance where it changed to None?? but whatever
                ]:
                    key_changes = diff.get(key)
                    if key_changes is None:
                        continue
                    (_prev_value, new_value) = key_changes
                    if new_value is not None:
                        # do not overwrite if new value is None
                        prev_row[key] = new_value
                    del diff[key]

                if len(diff) == 0:
                    continue

                # TODO check if there are any other unexpected diffs not in Status?
                status_dict = {key: new_value for (key, (_prev_value, new_value)) in diff.items()}
                prev_row.update(status_dict)
                status = Status(row=status_dict)
                ep.statuses.append(status)


# TODO how to make it hpi stat friendly?
# can still extract typing annotations etc, but I guess need to decide on some sort of convention?...
def processor() -> Processor:
    return Processor(config=make_config())



from my.tests.common import skip_if_not_karlicoss as pytestmark  # isort: skip

from more_itertools import ilen

# todo test against stats? not sure.. maybe both

def test_gdpr() -> None:
    import my.github.gdpr as gdpr

    assert ilen(gdpr.events()) > 100


def test() -> None:
    from my.github.all import get_events

    events = get_events()
    assert ilen(events) > 100
    for e in events:
        print(e)



'''
Just a demo module for testing and documentation purposes
'''
from __future__ import annotations

import json
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from datetime import datetime, timezone, tzinfo
from pathlib import Path
from typing import Protocol

from my.core import Json, PathIsh, Paths, get_files


class config(Protocol):
    data_path: Paths

    # this is to check required attribute handling
    username: str

    # this is to check optional attribute handling
    timezone: tzinfo = timezone.utc

    external: PathIsh | None = None

    @property
    def external_module(self):
        rpath = self.external
        if rpath is not None:
            from my.core.utils.imports import import_dir

            return import_dir(rpath)

        import my.config.repos.external as m  # type: ignore[import-not-found]

        return m


def make_config() -> config:
    from my.config import demo as user_config

    class combined_config(user_config, config): ...

    return combined_config()


@dataclass
class Item:
    '''
    Some completely arbitrary artificial stuff, just for testing
    '''

    username: str
    raw: Json
    dt: datetime


def inputs() -> Sequence[Path]:
    cfg = make_config()
    return get_files(cfg.data_path)


def items() -> Iterable[Item]:
    cfg = make_config()

    transform = (lambda i: i) if cfg.external is None else cfg.external_module.transform

    for f in inputs():
        dt = datetime.fromtimestamp(f.stat().st_mtime, tz=cfg.timezone)
        j = json.loads(f.read_text())
        for raw in j:
            yield Item(
                username=cfg.username,
                raw=transform(raw),
                dt=dt,
            )



"""
Bumble data from Android app database (in =/data/data/com.bumble.app/databases/ChatComDatabase=)
"""

from __future__ import annotations

import json
import sqlite3
from abc import abstractmethod
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Protocol, Union

from more_itertools import unique_everseen

from my.core import Paths, Res, get_files
from my.core.compat import assert_never
from my.core.sqlite import select, sqlite_connection


class Config(Protocol):
    @property
    @abstractmethod
    def export_path(self) -> Paths:
        """
        Paths[s]/glob to the exported sqlite databases
        """
        raise NotImplementedError

    @property
    def my_name(self) -> str:
        """
        Seems like there is no information about our own user in the database (not even name!).
        So if you want, you can supply the module with your name here.
        """
        return "me"


def make_config() -> Config:
    from my.config import bumble as user_config

    class combined_config(user_config.android, Config): ...

    return combined_config()


def inputs() -> Sequence[Path]:
    # TODO not ideal that we instantiate config here and in _entities...
    # perhaps should extract in a class (e.g. Processor or something like that)
    config = make_config()
    return get_files(config.export_path)


@dataclass(unsafe_hash=True)
class Person:
    user_id: str
    user_name: str


_MYSELF_USER_ID = '000000000'


# todo not sure about order of fields...
@dataclass
class _BaseMessage:
    id: str
    created: datetime
    is_incoming: bool
    text: str


@dataclass(unsafe_hash=True)
class _Message(_BaseMessage):
    conversation_id: str
    reply_to_id: str | None


@dataclass(unsafe_hash=True)
class Message(_BaseMessage):
    chat: Person
    sender: Person
    reply_to: Message | None


EntitiesRes = Res[Union[Person, _Message]]


def _entities() -> Iterator[EntitiesRes]:
    # Seems like there is no information about our own user in the database (not even name!), and there is no stable id we can rely on.
    # So we just make up a syncthetic user, and let the user specify the name in config
    config = make_config()

    yield Person(
        user_id=_MYSELF_USER_ID,
        user_name=config.my_name,
    )

    for db_file in inputs():
        with sqlite_connection(db_file, immutable=True) as db:
            yield from _handle_db(db)


def _handle_db(db: sqlite3.Connection) -> Iterator[EntitiesRes]:
    # todo hmm not sure
    # on the one hand kinda nice to use dataset..
    # on the other, it's somewhat of a complication, and
    # would be nice to have something type-directed for sql queries though
    # e.g. with typeddict or something, so the number of parameter to the sql query matches?
    # fmt: off
    for   user_id ,  user_name in select(
        ('user_id', 'user_name'),
        'FROM conversation_info',
        db=db,
    ):
    # fmt: on
        yield Person(
            user_id=user_id,
            user_name=user_name,
        )

    # NOTE
    # 'message' table:
    # - has sender_name, but it's always None
    # - has sender_id and recipient_id, but seems like they might change between app reinstalls?
    # - whereas conversation_id is surprisingly stable, e.g. even between account deletions/restores
    #     (checked on "Bumble" user)
    # fmt: off
    for  mid ,  conversation_id ,  created           ,  is_incoming ,  payload_type ,  payload ,  reply_to_id in select(
        ('id', 'conversation_id', 'created_timestamp', 'is_incoming', 'payload_type', 'payload', 'reply_to_id'),
        'FROM message ORDER BY created_timestamp',
        db=db,
    ):
    # fmt: on
        try:
            key = {'TEXT': 'text', 'QUESTION_GAME': 'text', 'IMAGE': 'url', 'GIF': 'url', 'AUDIO': 'url', 'VIDEO': 'url'}[payload_type]
            text = json.loads(payload)[key]
            yield _Message(
                id=mid,
                # TODO not sure if utc??
                created=datetime.fromtimestamp(created / 1000),
                is_incoming=bool(is_incoming),
                text=text,
                conversation_id=conversation_id,
                reply_to_id=reply_to_id,
            )
        except Exception as e:
            yield e


def _key(r: EntitiesRes):
    if isinstance(r, _Message):
        if '/hidden?' in r.text:
            # ugh. seems that image URLs change all the time in the db?
            # can't access them without login anyway
            # so use a different key for such messages
            # todo maybe normalize text instead? since it's gonna always trigger diffs down the line
            return (r.id, r.created)
    return r


_UNKNOWN_PERSON = "UNKNOWN_PERSON"


def messages() -> Iterator[Res[Message]]:
    id2person: dict[str, Person] = {}
    id2msg: dict[str, Message] = {}
    for x in unique_everseen(_entities(), key=_key):
        if isinstance(x, Exception):
            yield x
            continue
        if isinstance(x, Person):
            id2person[x.user_id] = x
            continue
        if isinstance(x, _Message):
            reply_to_id = x.reply_to_id
            # hmm seems that sometimes there are messages with no corresponding conversation_info?
            # possibly if user never clicked on conversation before..
            person = id2person.get(x.conversation_id)
            if person is None:
                person = Person(user_id=x.conversation_id, user_name=_UNKNOWN_PERSON)

            reply_to: Message | None = None
            if reply_to_id is not None:
                try:
                    reply_to = id2msg[reply_to_id]
                except Exception as e:
                    # defensive here, not a huge deal if we lost reply_to
                    yield e

            sender = person if x.is_incoming else id2person[_MYSELF_USER_ID]

            m = Message(
                id=x.id,
                created=x.created,
                # todo hmm is_incoming is a bit redundant?
                # think whether it can be useful in other providers or done in some generic way
                is_incoming=x.is_incoming,
                text=x.text,
                chat=person,
                sender=sender,
                reply_to=reply_to,
            )
            id2msg[m.id] = m
            yield m
            continue
        assert_never(x)
